# -*- coding: utf-8 -*-

"""Top-level package for aDMIX."""
__author__ = """Boris Bauermeister"""
__email__ = 'Boris.Bauermeister@gmail.com'
__version__ = '0.2.0'

#interfaces:
import os
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater


#tasks:
PKGDIR = os.path.dirname(__file__)

DEFAULT_CONFIG = os.path.join(PKGDIR, 'config', 'default.config')

from admix.download import download




#This class holds all information to read the pre-defined Rucio data
#format which is hold in a template configuration file. The Rucio naming
#convention is evaluated based on this input.

import json
import logging
#ToDo: Get your logging done in the right way!
#ToDo: Fix "public" vs. "private" declartions

class ConfigRucioDataFormat():
    #read/config rucio interface
    def __init__(self):
        self.rucio_configuration_ = None
        self.types_ = []
        self.structure_ = {}
        self.eval_ = 0

    def Config(self, config_path):
        #put in your favourite experimental configuration
        #to describe the rucio cataloge description
        self.config_path = config_path
        self.LoadConfig()
        self.Eval()

    def LoadConfig(self):
        with open(self.config_path) as f:
            self.rucio_configuration_ = json.load(f)

    def FindOccurrences(self, test_string, test_char):
        #https://stackoverflow.com/questions/13009675/find-all-the-occurrences-of-a-character-in-a-string
        return [i for i, letter in enumerate(test_string) if letter == test_char]

    def ExtractTagWords(self, test_string, beg_char, end_char):
        word_list = []
        char_beg = self.FindOccurrences(test_string, "{")
        char_end = self.FindOccurrences(test_string, "}")
        for j in range(len(char_beg)):
            j_char_beg = char_beg[j]
            j_char_end = char_end[j]
            word = test_string[j_char_beg+1:j_char_end]
            if word not in word_list:
                word_list.append(word)
        return word_list

    def Eval(self):
        #Get data types (=keys)
        self.types_ = list(self.rucio_configuration_.keys())

        if len(self.types_) == 0:
            print("Upload types are not defined")


        #Get the overall file structure from your configuration file (depends on experiment)
        self.structure_ = {}
        try:
            for i_type in self.types_:
                i_levels  = self.rucio_configuration_[i_type].split("|->|")
                nb_levels = len(i_levels)

                #init the level at least once beforehand
                self.structure_[i_type] = {}
                for idx, i_level in enumerate(i_levels):
                    self.structure_[i_type]["L"+str(idx)] = {}

                #fill the levels with information:
                for idx, i_level in enumerate(i_levels):
                    if i_level.find('$C') == 0:
                        self.structure_[i_type]["L"+str(idx)]['type'] = "rucio_container"
                        self.structure_[i_type]["L"+str(idx)]['did'] = i_level.replace("$C", "")
                    if i_level.find('$D') == 0:
                        self.structure_[i_type]["L"+str(idx)]['type'] = "rucio_dataset"
                        self.structure_[i_type]["L"+str(idx)]['did'] = i_level.replace("$D", "")

                    self.structure_[i_type]["L"+str(idx)]['tag_words'] = self.ExtractTagWords(i_level, "{", "}")
            self.eval_ = 1
        except:
            print("Evaluation failed")

    def GetTypes(self):
        return self.types_
    def GetStructure(self):
        return self.structure_
    def GetPlugin(self, plugin=None, reset=False):
        if reset == True:
            self.Eval()
        if plugin!=None and plugin in self.structure_:
            return self.structure_[plugin]
        else:
            return None



# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper
import time
import shutil

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater
from admix.utils import make_did

@Collector
class MoveDataToRSE():

    def __init__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')


        #Define data types
        self.NORECORDS_DTYPES = helper.get_hostconfig()['norecords_types']
        self.RAW_RECORDS_DTYPES = helper.get_hostconfig()['raw_records_types']
        self.RECORDS_DTYPES = helper.get_hostconfig()['records_types']

        #Get other parameters
        self.DATADIR = helper.get_hostconfig()['path_data_to_upload']
        self.periodic_check = helper.get_hostconfig()['upload_periodic_check']
        self.RSES = helper.get_hostconfig()['rses']

        self.FROMRSE = "UC_OSG_USERDISK"
        self.TORSE = "CCIN2P3_USERDISK"

        #Init the runDB
        self.db = ConnectMongoDB()

        #We want the first and the last run:
        self.gboundary = self.db.GetBoundary()
        self.run_nb_min = self.gboundary['min_number']
        self.run_nb_max = self.gboundary['max_number']
        self.run_ts_min = self.gboundary['min_start_time']
        self.run_ts_max = self.gboundary['max_start_time']

        #Init the Rucio data format evaluator in three steps:
        self.rc_reader = ConfigRucioDataFormat()
        self.rc_reader.Config(helper.get_hostconfig('rucio_template'))

        #This class will evaluate your destinations:
        self.destination = Destination()

        #Since we deal with an experiment, everything is predefine:
        self.exp_temp = Templater()
        self.exp_temp.Config(helper.get_hostconfig()['template'])

        #Init a class to handle keyword strings:
        self.keyw = Keyword()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()
        self.rc.SetProxyTicket("rucio_x509")
#        print(self.rc.Whoami())


    def add_conditional_rule(self,run_number, dtype, hash, from_rse, to_rse, lifetime=None, update_db=True):
        did = make_did(run_number, dtype, hash)
        result = self.rc.AddConditionalRule(did, from_rse, to_rse, lifetime=lifetime)
        #if result == 1:
        #   return
        helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}: conditional rule added: {2} ---> {3}'.format(run_number,dtype,did,to_rse))

        if update_db:
            rucio_rule = self.rc.GetRule(did, rse=to_rse)
            data_dict = {'host': "rucio-catalogue",
                         'type': dtype,
                         'location': to_rse,
                         'lifetime': rucio_rule['expires'],
                         'status': 'transferring',
                         'did': did,
                         'protocol': 'rucio'
                     }
            self.db.db.find_one_and_update({'number': run_number},
                                      {'$set': {'status': 'transferring'}}
                                  )

            docid = self.db.db.find_one({'number': run_number}, {'_id': 1})['_id']
            self.db.AddDatafield(docid, data_dict)




    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')


#        data_types = self.RAW_RECORDS_DTYPES + self.RECORDS_DTYPES + self.NORECORDS_DTYPES
        data_types = self.RAW_RECORDS_DTYPES + self.RECORDS_DTYPES


        # Get all runs that are already transferred and that still have some data_types in eb 
        cursor = self.db.db.find({
#            'number': {"$lt": 8000, "$gte": 7900},
            'number': {"$gte": 7300},
#            'number': 7597,
            'data' : { "$elemMatch": { "host" : "rucio-catalogue" , "type" : {"$in" : data_types}, "location" : self.FROMRSE }},
            'status': 'transferred'
        },
        {'_id': 1, 'number': 1, 'data': 1})

        cursor = list(cursor)

        helper.global_dictionary['logger'].Info('Runs that will be transferred are {0}'.format([c["number"] for c in cursor]))

        # Runs over all listed runs
        for run in cursor:
            number = run['number']
            helper.global_dictionary['logger'].Info('Treating run {0}'.format(number))
            for dtype in data_types:
                helper.global_dictionary['logger'].Info('\t==> Looking for data type {0}'.format(dtype))
                # get the datum for this datatype
                datum = None
                for d in run['data']:
                    if d['type'] == dtype and d['host'] == 'rucio-catalogue' and d['location'] == self.FROMRSE:
                        datum = d

                if datum is None:
                    helper.global_dictionary['logger'].Info('\t\t==> Data type {0} not in source RSE : {1}'.format(dtype,self.FROMRSE))
                    continue

                if 'did' not in datum:
                    helper.global_dictionary['logger'].Info('\t\t==> There is no did on the data type {1}'.format(dtype))
                    continue

                did = datum['did']

                hash = did.split('-')[-1]

                helper.global_dictionary['logger'].Info('\t\t==> Run {0}, data type {1}: found rule {2} in {3}'.format(number,dtype,did,self.FROMRSE))

                # check if a rule already exists in the destination RSE
                rucio_rule = self.rc.GetRule(upload_structure=did, rse=self.TORSE)
                if rucio_rule['exists'] and rucio_rule['state'] == 'OK':
                    helper.global_dictionary['logger'].Info('\t\t==> Data already in the destination RSE : {0}'.format(self.TORSE))
                    continue

                helper.global_dictionary['logger'].Info('\t\t==> Run {0}, data type {1}: rule added: {2} ---> {3}'.format(number,dtype,did,self.TORSE))

                self.add_conditional_rule(number, dtype, hash, self.FROMRSE, self.TORSE)


        return 0


    def __del__(self):
        pass



# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater

@Collector
class RunExampleTask():

    def __init__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')


        #Define data types
        self.DTYPES = helper.get_hostconfig()['rawtype']
        self.DATADIR = helper.get_hostconfig()['path_data_to_upload']
        self.periodic_check = helper.get_hostconfig()['upload_periodic_check']
#        self.DTYPES = ['raw_records', 'raw_records_he', 'raw_records_aqmon', 'raw_records_mv']
#        self.DATADIR = '/eb/ebdata'
#        self.periodic_check = 300

        #Init the runDB
        self.db = ConnectMongoDB()

        #We want the first and the last run:
        self.gboundary = self.db.GetBoundary()
        self.run_nb_min = self.gboundary['min_number']
        self.run_nb_max = self.gboundary['max_number']
        self.run_ts_min = self.gboundary['min_start_time']
        self.run_ts_max = self.gboundary['max_start_time']

        #Init the Rucio data format evaluator in three steps:
        self.rc_reader = ConfigRucioDataFormat()
        self.rc_reader.Config(helper.get_hostconfig('rucio_template'))

        #This class will evaluate your destinations:
        self.destination = Destination()

        #Since we deal with an experiment, everything is predefine:
        self.exp_temp = Templater()
        self.exp_temp.Config(helper.get_hostconfig()['template'])

        #Init a class to handle keyword strings:
        self.keyw = Keyword()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()
        self.rc.SetProxyTicket("rucio_x509")
#        print(self.rc.Whoami())


    def GetPathsFromRunNumber(self, run_number, dtype, hash):

        print("Run "+str(run_number)+", data type "+dtype+", hash "+hash)
        
        cursor = self.db.GetRunByNumber(run_number)        
        if len(cursor)==0:
            return
        run = cursor[0]

        file_replicas = {}

        for d in run['data']:
            if d['type'] != dtype:
                continue
            if d['host'] != 'rucio-catalogue':
                continue
            if d['status'] != 'transferred':
                continue
#            if d['location'] != 'LNGS_USERDISK': #'UC_DALI_USERDISK':
            if d['location'] != 'UC_DALI_USERDISK':
                continue
            did = d['did']
            if did.split('-')[-1] != hash:
                continue
            if self.rc.CheckRule(did, d['location']) != 'OK':
                continue
            print("Found",did)
            file_replicas = self.rc.ListFileReplicas(did,d['location'],localpath=True)

        return list(file_replicas.values())


    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')

        paths = self.GetPathsFromRunNumber(7177,"raw_records","rfzvpzj4mf")
        print(paths)


        return 0

    def __del__(self):
        """Use this function to clean up your class settings if necessary"""

#        print('dummy stop')
        pass



#Author: Boris Bauermeister
#Descriptions: Interface class to evaluate the destination

#import sys
#import datetime
#import os
import re

class Destination():

    def __init__(self):
        self._destination = None #holds the information about what is to do at the destination
        self._host        = None #holds the host where aDMIX is running
        self._origin      = None #holds the information from which locations the data destination

        self._rucio_rule  = []
        self._template = {
                            "rucio":"^rucio-catalogue:(\w+):(\w+)",
                            "rucio-download":"(\w+)",
                            "file":"^file:(\w+):(\w+)",
                         }


    def RunDestination(self):
        #print(self._host, self._origin, self._destination)

        self.__host_is_origin()
        self.__prepare_destination()

        #print(self._rucio_rule, self._host_is_origin)

        if self._host_is_origin == True and len(self._rucio_rule) > 0:
            #you are up for rucio uploads based on:
            # - your destination
            # - your host
            # --> upload host to destination with possible rules!
            return self._rucio_rule

        else:
            return []

    def EvalDestination(self, host, origin, destination):
        self._destination = destination
        self._origin      = origin
        self._host        = host
        self._rucio_rule  = []

        return self.RunDestination()

    def __prepare_destination(self):
        #make a list out of the destination just in case:
        if isinstance(self._destination, list) == False:
            self._destination = [self._destination]


        #evaluate the list of destinations:
        for i_dest in self._destination:
            #you might pass a None destination at
            #some point. But you want to ignore this
            if i_dest == None:
                continue
            if bool(re.search(self._template['rucio'], i_dest)):
                rc_dict={}
                rc_dict['protocol']='rucio-catalogue'
                rc_dict['rse'], rc_dict['lifetime'] = re.search(self._template['rucio'], i_dest).group(1,2)
                if len(self._rucio_rule) == 0:
                    rc_dict['upload'] = True
                else:
                    rc_dict['upload'] = False
                self._rucio_rule.append(rc_dict)
            if bool(re.search(self._template['file'], i_dest)):
                print(i_dest)
                rc_dict = {}
                rc_dict['protocol'] = 'file'
                rc_dict['dir'], rc_dict['info'] = re.search(self._template['file'], i_dest).group(1, 2)
                #if len(self._rucio_rule) == 0:
                #    rc_dict['upload'] = True
                #else:
                #    rc_dict['upload'] = False
                self._rucio_rule.append(rc_dict)

    def __host_is_origin(self):
        self._host_is_origin = False
        if self._host == self._origin:
            self._host_is_origin = True




import logging
import logging.handlers
import sys
from admix.helper.helper import global_dictionary, get_hostconfig
from admix import __version__ as admix_version

class Logger():

    def __init__(self,
                 logpath=None,
                 loglevel=logging.INFO,
                 ):

        if logpath != None:
            self.logpath = logpath
        else:
            self.logpath = "./"
        self.loglevel = loglevel
        self._loghandler = None
        self.setup()

    def setup(self):
        """Setup your logging handler is crucial

        """
        log_format = f'admix_v{admix_version} ' + '%(asctime)s [%(levelname)s]\t| %(message)s'

        self._loghandler = logging.getLogger('admix_logger')
        self._loghandler.setLevel(logging.DEBUG)
        # create file handler which logs even debug messages
        fh = logging.FileHandler(self.logpath)
        fh.setLevel(logging.DEBUG)
        # create console handler with a higher log level
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        # create formatter and add it to the handlers
        formatter = logging.Formatter(log_format)
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        # add the handlers to the logger
        self._loghandler.addHandler(fh)
        self._loghandler.addHandler(ch)

    def Info(self, info):
        self._loghandler.info(info)
    def Warning(self, warning):
        self._loghandler.warning(warning)
    def Debug(self, debug):
        self._loghandler.debug(debug)
    def Error(self, error):
        self._loghandler.error(error)



# -*- coding: utf-8 -*-
from admix.helper import helper

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater

@Collector
class InitTransfersMongoDB():

    def __init__(self):
        pass
    def __del__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')

        self.db = ConnectMongoDB()
        self.db.Connect()

        # We want the first and the last run:
        self.gboundary = self.db.GetBoundary()
        self.run_nb_min = self.gboundary['min_number']
        self.run_nb_max = self.gboundary['max_number']
        self.run_ts_min = self.gboundary['min_start_time']
        self.run_ts_max = self.gboundary['max_start_time']

        # Init the Rucio data format evaluator in three steps:
        self.rc_reader = ConfigRucioDataFormat()
        self.rc_reader.Config(helper.get_hostconfig('rucio_template'))

        # This class will evaluate your destinations:
        self.destination = Destination()

        # Since we deal with an experiment, everything is predefine:
        self.exp_temp = Templater()
        self.exp_temp.Config(helper.get_hostconfig()['template'])

        # Init a class to handle keyword strings:
        self.keyw = Keyword()

        # Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()


    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')

        ts_beg = None
        ts_end = None
        if helper.global_dictionary.get('run_numbers') != None:
            # Evaluate terminal input for run number assumption (terminal input == true)
            true_nb_beg, true_nb_end = helper.eval_run_numbers(helper.global_dictionary.get('run_numbers'),
                                                               self.run_nb_min,
                                                               self.run_nb_max)
            # Get the timestamps from the run numbers:
            ts_beg = self.db.FindTimeStamp('number', int(true_nb_beg))
            ts_end = self.db.FindTimeStamp('number', int(true_nb_end))

        elif helper.global_dictionary.get('run_timestamps') != None:
            # Evaluate terminal input for run name assumption
            true_ts_beg, true_ts_end = helper.eval_run_timestamps(helper.global_dictionary.get('run_timestamps'),
                                                                  self.run_ts_min,
                                                                  self.run_ts_max)
            ts_beg = true_ts_beg
            ts_end = true_ts_end

        elif helper.global_dictionary.get('run_timestamps') == None and \
                helper.global_dictionary.get('run_numbers') == None:
            ts_beg = self.run_ts_min
            ts_end = self.run_ts_max
        else:
            helper.global_dictionary['logger'].Error(
                "Check for your input arguments (--select-run-number or --select-run-time")
            exit(1)
            # exection

        # After we know the times:
        helper.global_dictionary['logger'].Info(f"Run between {ts_beg} and {ts_end}")

        #Get your collection of run numbers and run names
        collection = self.db.GetDestination(ts_beg, ts_end)

        #Run through the overview collection:
        for i_run in collection:
            #Extract run number and name from overview collection
            r_name = i_run['name']

            dict_name = {}
            dict_name['date'] = r_name.split("_")[0]
            dict_name['time'] = r_name.split("_")[1]

            r_number = i_run['number']
            #Pull the full run information (according to projection which is pre-defined) by the run name
            db_info = self.db.GetRunByName(r_name)[0]

            #REMOVE AND CHECK FOR UPLOAD DETECTORS (-> config files)
            #ToDo: Make it work with XENONnT later
            if db_info['detector'] != 'tpc':
                continue

            for i_data in db_info['data']:

                #get the destination from DB:
                origin_dest = None
                if 'destination' in i_data:
                    origin_dest = i_data['destination']

                #build a destination from host=origin (=rucio-catalogue)
                dest = self.destination.EvalDestination(host="rucio-catalogue",
                                                        origin=i_data['host'],
                                                        destination=origin_dest)

                if len(dest) == 0:
                    helper.global_dictionary['logger'].Info("No destination specified! Skip")
                    continue

                #At this stage we know of an existing destination:
                rc_scope = i_data['location'].split(":")[0]
                rc_dname = i_data['location'].split(":")[1]

                rucio_template = self.rc_reader.GetPlugin(i_data['type'])
                rucio_template_sorted = [key for key in sorted(rucio_template.keys())]

                # Fill the key word class with information beforehand:
                self.keyw.ResetTemplate()
                #self.keyw.SetTemplate(template_info)
                self.keyw.SetTemplate(db_info)
                self.keyw.SetTemplate({'plugin': i_data['type'],
                                       "hash": i_data['location'].split(":")[-1].split("-")[-1]
                                       })
                self.keyw.SetTemplate(dict_name)
                self.keyw.SetTemplate({'science_run': helper.get_science_run(db_info['start'])})

                rucio_template = self.keyw.CompleteTemplate(rucio_template)

                helper.global_dictionary['logger'].Info('Type {0}'.format(i_data['type']))
                helper.global_dictionary['logger'].Debug('Run is at host: {0}'.format(i_data['host']))
                helper.global_dictionary['logger'].Debug('aDMIX runs at host: {0}'.format(helper.get_hostconfig('host')))

                #Begin to work through the new/updated rucio destinations:
                for i_dest in dest:
                    #print(i_dest)
                    rse_rules = ["{protocol}:{rse}:{lifetime}".format(protocol=i_dest['protocol'],
                                                                     rse=i_dest['rse'],
                                                                     lifetime=i_dest['lifetime'])]

                    try:
                        rule = self.rc.GetRule(upload_structure=rucio_template, rse=i_dest['rse'])
                    except:
                        helper.global_dictionary['logger'].Error("No rule transfer rule applied for {0}".format(i_dest['rse']))
                        continue
                    #1) Rule does not exists -> Create it, delete destination
                    if rule['state'] == "Unkown" and rule['rse'] == None:
                        try:
                            helper.global_dictionary['logger'].Info('Add a Rucio transfer rule')
                            self.rc.AddRules(upload_structure=rucio_template, rse_rules= rse_rules)
                        except:
                            helper.global_dictionary['logger'].Error('Can not create Rucio transfer rule')
                    #2) Rule exists but lifetime is different -> Update lifetime, delete destination
                    if rule['state'] == "OK" and i_dest['lifetime'] == 'None':
                        helper.global_dictionary['logger'].Info('No need to do anything!')
                    #3) Rule does exists without lifetime changes -> delete destination
                    if rule['state'] == "OK" and i_dest['lifetime'] != 'None':
                        try:
                            helper.global_dictionary['logger'].Info('Update Rucio transfer rule')
                            self.rc.UpdateRules(upload_structure=rucio_template, rse_rules= rse_rules)
                        except:
                            helper.global_dictionary['logger'].Error('Can not update a transfer rule')


                #update runDB after rules are updated/added:
                rucio_rules = self.rc.ListDidRules(rucio_template)

                rse_to_db = []
                for i_rule in rucio_rules:
                    if i_rule['expires_at'] == None:
                        dt = None
                    else:
                        dt = i_rule['expires_at'].strftime("%Y-%m-%d-%H:%M:%S")
                    db_field = "{rse}:{state}:{lifetime}".format(rse=i_rule['rse_expression'],
                                                                 state=i_rule['state'],
                                                                 lifetime=dt)
                    rse_to_db.append(db_field)

                helper.global_dictionary['logger'].Info(f'Update database with {rse_to_db}')
                self.db.SetDataField(db_info['_id'], type=i_data['type'],
                                                     host='rucio-catalogue',
                                                     key='rse',
                                                     value=rse_to_db)

                #Last step is to evaluate if InitTransfersWithMongoDB was able to init new Rucio transfer rules
                #extract a list of required destinations:
                t_dest = [ i['rse'] for i in dest ]

                for i_rule_found in rse_to_db:
                    #information about rules which are found in Rucio about the DID
                    t_rse    = i_rule_found.split(":")[0]


                    if t_rse in t_dest:
                        dest[:] = [f"{i_dict.get('protocol')}:{i_dict.get('rse')}:{i_dict.get('lifetime')}" for
                                       i_dict in dest if i_dict.get('rse') != t_rse]

                        helper.global_dictionary['logger'].Info(f'Update database with destinations {dest}')

                        self.db.SetDataField(db_info['_id'], type=i_data['type'],
                                             host='rucio-catalogue',
                                             key='destination',
                                             value=dest)
                    else:
                        helper.global_dictionary['logger'].Info(f'Database destinations are still: {dest}')




# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater

@Collector
class UploadMongoDB():

    def __init__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')

        #Init the runDB
        self.db = ConnectMongoDB()
        self.db.Connect()

        #We want the first and the last run:
        self.gboundary = self.db.GetBoundary()
        self.run_nb_min = self.gboundary['min_number']
        self.run_nb_max = self.gboundary['max_number']
        self.run_ts_min = self.gboundary['min_start_time']
        self.run_ts_max = self.gboundary['max_start_time']

        #Init the Rucio data format evaluator in three steps:
        self.rc_reader = ConfigRucioDataFormat()
        self.rc_reader.Config(helper.get_hostconfig('rucio_template'))

        #This class will evaluate your destinations:
        self.destination = Destination()

        #Since we deal with an experiment, everything is predefine:
        self.exp_temp = Templater()
        self.exp_temp.Config(helper.get_hostconfig()['template'])

        #Init a class to handle keyword strings:
        self.keyw = Keyword()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()

    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')


        ts_beg = None
        ts_end = None
        if helper.global_dictionary.get('run_numbers') != None:
            #Evaluate terminal input for run number assumption (terminal input == true)
            true_nb_beg, true_nb_end = helper.eval_run_numbers(helper.global_dictionary.get('run_numbers'),
                                                               self.run_nb_min,
                                                               self.run_nb_max)
            #Get the timestamps from the run numbers:
            ts_beg = self.db.FindTimeStamp('number', int(true_nb_beg))
            ts_end = self.db.FindTimeStamp('number', int(true_nb_end))

        elif helper.global_dictionary.get('run_timestamps') != None:
            #Evaluate terminal input for run name assumption
            true_ts_beg, true_ts_end = helper.eval_run_timestamps(helper.global_dictionary.get('run_timestamps'),
                                                               self.run_ts_min,
                                                               self.run_ts_max)
            ts_beg = true_ts_beg
            ts_end = true_ts_end

        elif helper.global_dictionary.get('run_timestamps') == None and \
            helper.global_dictionary.get('run_numbers') == None:
            ts_beg = self.run_ts_min
            ts_end = self.run_ts_max
        else:
            helper.global_dictionary['logger'].Error("Check for your input arguments (--select-run-number or --select-run-time")
            exit(1)
            #exection

        #After we know the times:
        helper.global_dictionary['logger'].Info(f"Run between {ts_beg} and {ts_end}")

        #Get your collection of run numbers and run names
        collection = self.db.GetDestination(ts_beg, ts_end)

        print(collection)

        #Run through the overview collection:
        for i_run in collection:

            #Extract run number and name from overview collection
            r_name = i_run['name']
            r_number = i_run['number']
            #print("test")

            #Pull the full run information (according to projection which is pre-defined) by the run name
            db_info = self.db.GetRunByName(r_name)[0]

            #REMOVE AND CHECK FOR UPLOAD DETECTORS (-> config files)
            #ToDo: Make it work with XENONnT later
            if db_info['detector'] != 'tpc':
                continue

            helper.global_dictionary['logger'].Info(f"Run: {r_name} / {r_number}")

            for i_data in db_info['data']:

                #get the destination from DB:
                origin_dest = None
                if 'destination' in i_data:
                    origin_dest = i_data['destination']

                dest = self.destination.EvalDestination(host=helper.get_hostconfig('host'),
                                                        origin=i_data['host'],
                                                        destination=origin_dest)

                if len(dest) == 0:
                    helper.global_dictionary['logger'].Debug("No destination specified! Skip")
                    continue

                helper.global_dictionary['logger'].Info('Type {0}'.format(i_data['type']))
                helper.global_dictionary['logger'].Debug('Run is at host: {0}'.format(i_data['host']))
                helper.global_dictionary['logger'].Debug('aDMIX runs at host: {0}'.format(helper.get_hostconfig('host')))

                #reactivate if this is specified in the config file
                #if helper.global_dictionary['plugin_type'] != None and helper.global_dictionary['plugin_type'] != i_date['type']:
                    #continue

                ptr0 = None
                rlt0 = None
                rse0 = None
                for i_dest in dest:
                    if i_dest['upload'] == True:
                        ptr0 = i_dest['protocol']
                        rlt0 = i_dest['lifetime']
                        if rlt0 == "None":
                            rlt0 = None
                        rse0 = i_dest['rse']
                helper.global_dictionary['logger'].Debug("Evaluated destination for upload:")
                helper.global_dictionary['logger'].Debug(f" > Protocol: {ptr0}")
                helper.global_dictionary['logger'].Debug(f" > Lifetime: {rlt0}")
                helper.global_dictionary['logger'].Debug(f" > RSE: {rse0}")

                #pre infos:
                origin_type     = i_data['type'] #This is the plugin name which is gonna be handled (uploaded)
                origin_location = i_data['location'] #This holds the physical location of the plugin folder
                origin_host     = i_data['host']
                origin_status   = i_data['status']

                helper.global_dictionary['logger'].Debug("Evaluate origin information:")
                helper.global_dictionary['logger'].Debug(f" > Origin Host: {origin_host}")
                helper.global_dictionary['logger'].Debug(f" > Origin Status: {origin_status}")

                #Extract the template information according the pre-defined physical file structure:
                template_info = self.exp_temp.GetTemplateEval(plugin=origin_type,
                                                              host=origin_host,
                                                              string=origin_location)


                #print("template info:")
                #print(template_info)

                #print("rucio plugin info:")
                #print(":", self.rc_reader.GetPlugin(origin_type, reset=True))
                #print(":", self.rc_reader.GetTypes())

                #Evaluate the Rucio name template according the plugin which is requested:
                # rucio_template: Holds the unsorted Rucio container/dataset/file structure in levels
                # rucio_template_sorted: Holds the sorted levels (L0, L1, L2,...)
                rucio_template = self.rc_reader.GetPlugin(origin_type, reset=True)
                rucio_template_sorted = [key for key in sorted(rucio_template.keys())]


                #Fill the key word class with information beforehand:
                self.keyw.ResetTemplate()
                self.keyw.SetTemplate(template_info)
                self.keyw.SetTemplate(db_info)
                self.keyw.SetTemplate({'science_run': helper.get_science_run(db_info['start'])})

                rucio_template = self.keyw.CompleteTemplate(rucio_template)

                #collect all files from the directory and do pre checks:
                if os.path.isdir(origin_location) == False:
                    helper.global_dictionary['logger'].Error(f"Folder {origin_location} does not exists")
                    continue

                #load files:
                r_path, r_folder, r_files = helper.read_folder(origin_location)

                #check if files are there:
                if len(r_files) == 0:
                    helper.global_dictionary['logger'].Error(f"There are no files in the folder {origin_location}")
                    continue

                #check for zero sizes files before hand:
                file_checks = False
                for i_file in r_files:
                    f_size = os.path.getsize( os.path.join(origin_location, i_file) )
                    #print(i_file, f_size)
                    if f_size == 0:
                        helper.global_dictionary['logger'].Error(f"File {i_file} is 0 bytes")
                        helper.global_dictionary['logger'].Error(f"Exclude directory {origin_location} from upload")
                        file_checks=True
                        continue
                    if '_temp' in i_file:
                        #avoid to upload temp. incomplete files (ToDo need to checked with Strax!)
                        helper.global_dictionary['logger'].Error(f"Identified _temp extension in file name {i_file}")
                        file_checks=True
                        continue
                    if template_info['plugin'] not in i_file or template_info['hash'] not in i_file:
                        #avoid to upload wrong file names. (ToDo: Maybe we can improve in this test but works for not)
                        helper.global_dictionary['logger'].Error(f"Plugin name and hash is not recognized in file name {i_file}")
                        file_checks=True
                        continue

                if file_checks == True:
                    helper.global_dictionary['logger'].Info(f"File checks not successful!")
                    continue

                #Check if rule exists in Rucio:
                rule_status = self.rc.CheckRule(upload_structure=rucio_template, rse=rse0)
                #verify the locations with the rucio catalogue
                #Hint: the outcome of this check does not skip uploading! (missing files on disk):
                verify_location, _, _ = self.rc.VerifyLocations(upload_structure=rucio_template, upload_path=origin_location, checksum_test=False)
                #check for a database entry:
                db_dest_status = self.db.StatusDataField(db_info['_id'], type=origin_type, host=ptr0)

                helper.global_dictionary['logger'].Info(f"Rucio rule status is: {rule_status}")
                helper.global_dictionary['logger'].Info(f"Rucio vs. Disk file verfication is: {verify_location}")
                helper.global_dictionary['logger'].Info(f"Database status is: {db_dest_status}")

                #logging the filled template for debugging:
                for tmpl_key, tmpl_val in rucio_template.items():
                    helper.global_dictionary['logger'].Debug(f"Level: {tmpl_key}")
                    for inn_key, inn_val in tmpl_val.items():
                        helper.global_dictionary['logger'].Debug(f"  - {inn_key} : {inn_val}")

#                p = self.db.GetDataField(db_info['_id'], type=origin_type, host=ptr0, key='status')
#                print(p, origin_type)
#                #self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='status', value='RSEreupload')
#                continue

                skip_upload = False
                if db_dest_status == False:
                    #Create a new dictionary in the data field for the plugin/host combination
                    new_data_dict={}
                    new_data_dict['location'] = "n/a"
                    new_data_dict['status'] = "transferring"
                    new_data_dict['host'] = ptr0
                    new_data_dict['type'] = origin_type
                    new_data_dict['meta'] = None
                    if ptr0 == 'rucio-catalogue':
                        new_data_dict['rse'] = []
                        new_data_dict['destination'] = []

                    self.db.AddDatafield(db_info['_id'], new_data_dict)

                elif rule_status == 'OK':
                    #if there is a rucio rule we can skip here
                    skip_upload = True
                    helper.global_dictionary['logger'].Info("A Rucio rule exits already. We skip here!")

                elif db_dest_status == True:

                    if self.db.GetDataField(db_info['_id'],
                                            type=origin_type,
                                            host=ptr0, key='status') == 'transferring' or \
                       self.db.GetDataField(db_info['_id'],
                                             type=origin_type,
                                             host=ptr0, key='status') == 'transferred' or \
                        self.db.GetDataField(db_info['_id'],
                                                 type=origin_type,
                                                 host=ptr0, key='status') == 'RucioClearance':

                        skip_upload = True
                    elif self.db.GetDataField(db_info['_id'],
                                            type=origin_type,
                                            host=ptr0, key='status') == 'RSEreupload':
                        skip_upload = False

                if helper.global_dictionary['force'] == True:
                    helper.global_dictionary['logger'].Info("Enable forced upload to Rucio")
                    skip_upload = False


                #elif db_dest_status == True and self.db.GetDataField(db_info['_id'], type=origin_type, host=ptr0, key='status') == 'RSEreupload':
                #    #There is a database status found in RSEreupload state: Reset it and continoue with uploading:
                #    skip_upload = False
                #    #self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='status', value='transferring')
                #    #self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='location', value='n/a')
                #elif db_dest_status == True and len(self.db.GetDataField(db_info['_id'], type=origin_type, host=ptr0, key='rse'))>0:
                #    #if there exists a list of RSEs for the rucio-catalogue entry which contains the upload RSE we skip
                #    list_rse_is = self.db.GetDataField(db_info['_id'], type=origin_type, host=ptr0, key='rse')
                #    for i_rse_is in list_rse_is:
                #        print("-", i_rse_is)
                #        if i_rse_is.split(":")[0]==rse0 and i_rse_is.split(":")[1]=='OK':
                #            skip_upload = True

                if skip_upload==True:
                    tmp_status = self.db.GetDataField(db_info['_id'], type=origin_type, host=ptr0, key='status')
                    tmp_rse = self.db.GetDataField(db_info['_id'], type=origin_type, host=ptr0, key='rse')
                    helper.global_dictionary['logger'].Info("We skip uploading here because:\n "\
                                                            f"Database status is {tmp_status} and "\
                                                            f"Database RSE is {tmp_rse}. "\
                                                            "We do not want to interfere here...")
                    continue



                #if the pre checks are ok we can upload:
                upload_result = 1

                try:
                    self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='status', value='transferring')
                    helper.global_dictionary['logger'].Info("Start Rucio upload...")
                    upload_result = self.rc.Upload(upload_structure=rucio_template,
                                                   upload_path=origin_location,
                                                   rse=rse0,
                                                   rse_lifetime=rlt0)
                    helper.global_dictionary['logger'].Info("Rucio upload finished")
                except:
                    helper.global_dictionary['logger'].Error("Rucio upload failed")
                    #We can skip everything after the rucio upload failed
                    #Just set the Rucio database entry into RSEreupload to allow admix
                    #to try it again later
                    self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='status', value='RSEreupload')
                    continue

                #Intend that the try/execpt run through well but there was a problem with the upload/rule setting
                #of the DID. In that case we do not want to contioue:
                if upload_result == 1:
                    helper.global_dictionary['logger'].Error("Rucio upload may have failed. Check for DID")
                    #continue


                #Verify if all files are uploaded (verify_location == True is enough)
                verify_location, _, _ = self.rc.VerifyLocations(upload_structure=rucio_template, upload_path=origin_location, checksum_test=False)
                rucio_rule = self.rc.GetRule(upload_structure=rucio_template, rse=rse0)

                ##update after initial upload:
                if verify_location == True and rucio_rule['state'] == 'OK':
                    helper.global_dictionary['logger'].Info("Rucio rule status after upload is OK")

                    #get & set rse information and location:
                    rse_to_db = ["{rse}:{state}:{lifetime}".format(rse=rse0, state=rucio_rule['state'], lifetime=rucio_rule['expires'])]
                    rc_location = rucio_template[rucio_template_sorted[-1]]['did']

                    self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='rse', value=rse_to_db)
                    self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='location', value=rc_location)
                    self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='status', value='transferred')

                    helper.global_dictionary['logger'].Info(f"RSE information: {rse_to_db}")
                    helper.global_dictionary['logger'].Info(f"DID location: {rc_location}")
                    helper.global_dictionary['logger'].Info("Switch DB status to transferred")


                    #if everything was ok we can dump the metadata.json to the runDB:
                    meta_data_file = None
                    for i_file in r_files:
                        if 'metadata.json' in i_file:
                            meta_data_file = os.path.join(origin_location, i_file)
                            break
                    if os.path.exists(meta_data_file):
                        md = json.load(open(meta_data_file))
                        helper.global_dictionary['logger'].Info(f"Metadata file {meta_data_file} found and loaded")
                    else:
                        helper.global_dictionary['logger'].Error(f"Metadata file {meta_data_file} not found!")
                        md = {}
                    self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='meta', value=md)


                    #Clean up: Remove destination field and add some rules:
                    #remove destination entry:
                    #1 find out if other destinations are on the list:
                    for i_dest in dest:
                        if i_dest['upload'] == False:
                            ptr1 = i_dest['protocol']
                            rlt1 = i_dest['lifetime']
                            rse1 = i_dest['rse']

                            #apply further rucio rules
                            if ptr1 == 'rucio-catalogue':
                                rule_list = ["{ptr}:{rse}:{rlt}".format(ptr=ptr1,
                                                                        rse=rse1,
                                                                        rlt=rlt1)]
                                try:
                                    helper.global_dictionary['logger'].Info(f"Apply {rule_list} to Rucio DID")
                                    self.rc.AddRules(upload_structure=rucio_template, rse_rules=rule_list)
                                except:
                                    helper.global_dictionary['logger'].Error(f"Rule {rule_list} can not be applied!")

                    #if we have a complete uploaded data set and further rules initiated
                    #we can update the destination list:
                    #create an empty list of rule status messages:
                    rucio_rule_status = []

                    #get rule for the uploaded location:
                    try:
                        rucio_rule_rse0 = self.rc.GetRule(upload_structure=rucio_template, rse=rse0)
                        rucio_rule_status.append("{rse}:{state}:{lifetime}".format(rse=rse0,
                                                                                   state=rucio_rule_rse0['state'],
                                                                                   lifetime=rucio_rule_rse0['expires']))
                    except:
                        helper.global_dictionary['logger'].Error(f"Get Rucio rule for {rse0} failed"\
                                                                 f"We skip further Rucio rule requests"\
                                                                 f"and leave it to the admix rule updater")
                    #collect more rule status information about the
                    for i_dest in dest:
                        if i_dest['upload'] == False:
                            ptr1 = i_dest['protocol']
                            rlt1 = i_dest['lifetime']
                            rse1 = i_dest['rse']

                            #apply further rucio rules
                            if ptr1 == 'rucio-catalogue':
                                try:
                                    rucio_rule_rse1 = self.rc.GetRule(upload_structure=rucio_template, rse=rse1)
                                    rucio_rule_status.append("{rse}:{state}:{lifetime}".format(rse=rse1,
                                                                                           state=rucio_rule_rse1['state'],
                                                                                           lifetime=rucio_rule_rse1['expires']))
                                except:
                                    helper.global_dictionary['logger'].Error(f"Get Rucio rule for {rse0} failed "\
                                                                             f"We skip further Rucio rule requests "\
                                                                             f"and leave it to the admix rule updater")

                    #delete the destination if everything seems ok:
                    if rucio_rule_rse0['state'] == 'OK' and len(rucio_rule_status) == len(dest):
                        helper.global_dictionary['logger'].Info("Upload and transfer rule setting was SUCCESSFUL\n}"
                                                                "-> Empty the destination field at the host")
                        self.db.SetDataField(db_info['_id'], type=origin_type, host=origin_host, key='destination', value=[])

                    #But also update the data field based on the latest rucio information:
                    helper.global_dictionary['logger'].Info("Finally: Update Rucio entry in the database by:")
                    for i_rule in rucio_rule_status:
                        helper.global_dictionary['logger'].Info(f" > Rule: {i_rule}")

                    self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='rse', value=rucio_rule_status)

                else:
                    #rse_to_db = ["{rse}:{state}:{lifetime}".format(rse=rse0,
                    #                                               state=rucio_rule['state'],
                    #                                               lifetime=rucio_rule['expires'])]
                    #self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='rse', value=rse_to_db)
                    helper.global_dictionary['logger'].Error("Something went wrong during the upload and we can not"\
                                                             "verify the Rucio transfer rule status"\
                                                             "-> Set to RSEreupload in the database.")
                    self.db.SetDataField(db_info['_id'], type=origin_type, host=ptr0, key='status', value='RSEreupload')

        return 0


    def __del__(self):
        pass



import pymongo
from utilix.rundb import pymongo_collection
import admix.helper.helper as helper
import datetime


class ConnectMongoDB():

    def __init__(self):

        # instantiates self.db
        # do it this way so that we can use this method to reconnect if needed
        self.Connect()

        #Define the basic projections for mongoDB
        self.SetProjection(projection=None, from_config=True)


    def SetProjection(self, projection=None, from_config=False):

        if from_config == True:
            if 'projection' not in helper.get_hostconfig()['database']:
                self.projection = None
            else:
                self.projection = helper.get_hostconfig()['database']['projection']
        else:
            self.projection = projection


    def Connect(self):
        # nT runs DB
        self.db = pymongo_collection('runs')

        # for querying the hash collection for strax data types
        self.contexts = pymongo_collection('contexts')

    def GetQuery(self, query, reconnect=False, sort=[('_id',1)]):
        if reconnect == True:
            self.Connect()

        return list(self.db.find(query, projection=self.projection).sort(sort) )


    def GetRunByNumber(self, run_number, reconnect=False):

        if reconnect == True:
            self.Connect()

        query = {"number": run_number}
        return list(self.db.find(query, projection=self.projection ))


    def GetRunByName(self, run_name, reconnect=False):
        if reconnect == True:
            self.Connect()

        query = {"name": run_name}
        return list(self.db.find(query, projection=self.projection ))

    def GetRunByNameNumber(self, run_name, run_number, reconnect=False):
        if reconnect == True:
            self.Connect()

        query = { "$and":[ {"name": run_name}, {"number":run_number}]}
        return list(self.db.find(query, projection=self.projection ))

    def GetRunByID(self, run_id, reconnect=False):
        if reconnect == True:
            self.Connect()

        query = {"_id": run_id}
        return list(self.db.find(query, projection=self.projection ))

    def GetRunsByTimestamp(self, ts_beg=None, ts_end=None, reconnect=False, sort=[('_id',1)]):
        if reconnect == True:
            self.Connect()

        if ts_beg == None:
            ts_beg = self.GetSmallest("start")
        if ts_end == None:
            ts_end = self.GetLargest("start")

        query = { '$and': [ {'start': {'$gte':ts_beg}}, {'start': {'$lte':ts_end}} ] }
        projection = {"name":True, "number": True, "start":True, "_id":True}
        return list(self.db.find(query, projection=projection).sort(sort) )

    def GetSmallest(self, key="number"):
        return sorted(list(self.db.find({}, projection={"name": True, "number":True, "_id":True})), key=lambda k: k["number"])[0].get(key)

    def GetLargest(self, key="number"):
        return sorted(list(self.db.find({}, projection={"name": True, "number":True, "_id":True})), key=lambda k: k["number"])[-1].get(key)
        #return sorted(list(self.db.find({}, projection={key: True, '_id':True})), key=lambda k: k[key])[-1][key]

    def GetBoundary(self, key="number"):
        #Evaluate the full run database for the first and last element and return
        #a dictionary with its information. If None, then db entry does not exists
        klist = sorted(list(self.db.find({}, projection={"name": True, "number":True, "start": True, "_id":True})), key=lambda k: k["number"])
        kdict = {"min_name": klist[0].get('name'),
                 "min_number": klist[0].get('number'),
                 "min_start_time": klist[0].get('start'),
                 "max_name": klist[-1].get('name'),
                 "max_number": klist[-1].get('number'),
                 "max_start_time": klist[-1].get('start'),
                 }
        return kdict


    def GetRunsByTag(self, tag=None, sort="ascending"):

        #check if tag is a list or a string:
        if type(tag) == str:
            tag = [tag]
        elif type(tag) != list or tag==None:
            return {}

        #create an aggregation from the list of tags:
        tag_name_match = [ {"tags.name": i_tag} for i_tag in tag]
        tag_name_match = {"$match": {"$and": tag_name_match}}

        #evaluate sort direction:
        if sort == "ascending":
            sort = -1 #ascending means from latest to earliest run
        else:
            sort = 1

        agr = [
            {"$project":  {"tags": True, "name": True, "number": True, "start":True}},
            tag_name_match,
            {"$sort": {"start": sort}}
            ]
        return list(self.db.aggregate(agr))

    def GetRunsBySource(self, source=None, sort="ascending"):

        #check if source is a list or a string:
        if type(source) == str:
            source = [source]
        elif type(source) != list or source==None:
            return {}

        #create an aggregation from the list of tags:
        source_name_match = [ {"source.type": i_source} for i_source in source]
        source_name_match = {"$match": {"$and": source_name_match}}

        #evaluate sort direction:
        if sort == "ascending":
            sort = -1 #ascending means from latest to earliest run
        else:
            sort = 1

        agr = [
            {"$project":  {"name": True, "number": True, "start":True, "source":True}},
            source_name_match,
            {"$sort": {"start": sort}}
            ]
        return list(self.db.aggregate(agr))

    def FindTimeStamp(self, key, value):
        #timestamp definition:
        # The timestamp is defined according the 'start' tag in the runDB:
        runDB_tag_timestamp = 'start'
        # <Hard coded Change later>
        query = {key:value}
        return list(self.db.find(query, projection={key: True, '_id':True, runDB_tag_timestamp:True}))[0][runDB_tag_timestamp]


    def SetDestination(self, id_field, type=None, host=None, destination=None):
        #This function is database entry specific:
        #It changes a list (data) which contains dictionaries
        #In particular you can change here the destination field
        run = self.GetRunByID(id_field)[0]

        old_data = run['data']
        new_data = old_data.copy()

        if type != None and host!=None and destination != None:
            for i_run in new_data:
                if i_run['type'] != type:
                    continue
                if i_run['host'] != host:
                    continue
                if destination != None:
                    i_run['destination'] = destination

        self.db.find_one_and_update({"_id": id_field},
                                        {"$set": {"data": new_data}})

    def GetDataField(self, id_field, type=None, host=None,
                     key=None):

        run = self.GetRunByID(id_field)[0]

        value = "dict_not_exists"
        for i_run in run['data']:
            if i_run['type'] != type:
                continue
            if i_run['host'] != host:
                continue
            if key not in i_run:
                value="key_not_exists"
            elif key in i_run:
                value=i_run[key]
        #if value==dict_not_exists: given host/type combination in the dictionary does not exits
        #elif: key not in dict: key_not_exists
        #elif: return key of dict
        return value

    def GetDestination(self, ts_beg=None, ts_end=None, sort="ascending"):

        if ts_beg == None:
            ts_beg = self.GetSmallest("start")
        if ts_end == None:
            ts_end = self.GetLargest("start")

        #evaluate sort direction:
        if sort == "ascending":
            sort = -1 #ascending means from latest to earliest run
        else:
            sort = 1

        col = self.db.aggregate(
                    [
                        {"$match": {'$and': [ {'start': {'$gte':ts_beg}}, {'start': {'$lte':ts_end}} ]}},
                        {"$unwind": "$data"},
                        {"$project": {
                                        "name": 1, "number": 1, "destination": "$data.destination",
                                        "ndest": { "$cond": [{ "$isArray": "$data.destination"}, {"$size": "$data.destination"}, 0]},
                                    }},
                        {"$match": {"ndest": {"$gt": 0}}},
                        {"$project": {"name":1, "number":1}},
                        {"$sort": {"start": sort}}

                    ]
                )

        return col

    def GetDestinationTest(self, ts_beg=None, ts_end=None, sort="ascending"):

        if ts_beg == None:
            ts_beg = self.GetSmallest("start")
        if ts_end == None:
            ts_end = self.GetLargest("start")

        #evaluate sort direction:
        if sort == "ascending":
            sort = -1 #ascending means from latest to earliest run
        else:
            sort = 1

        col = self.db.aggregate(
                    [
                        {"$match": {'$and': [ {'start': {'$gte':ts_beg}}, {'start': {'$lte':ts_end}} ]}},
                        {"$unwind": "$data"},
                        {"$project": {
                                        "name": 1, "number": 1, "destination": "$data.location",
                                        "ndest": { "$cond": [{ "$isArray": "$data.location"}, {"$size": "$data.location"}, 0]},
                                    }},
#                        {"$match": {"ndest": {"$gt": 0}}},
                        {"$project": {"name":1, "number":1}},
                        {"$sort": {"start": sort}}

                    ]
                )

        return col


    def GetHosts(self, host, ts_beg=None, ts_end=None, sort="ascending"):
        """Function: GetLocations
        Get a selection of events between two timestamps (ts_beg, ts_end) for a pre-selected host
        in the meta database.
        :param host: A general location search to look for results of a specific host in the meta database
        :param ts_beg: datetime object when to begin the search
        :param ts_end: datetime object when to end the search
        :param sort: How to sort the result: ascending or descending
        :return MongoDB aggregation:
        """

        if ts_beg == None:
            ts_beg = self.GetSmallest("start")
        if ts_end == None:
            ts_end = self.GetLargest("start")

        #evaluate sort direction:
        if sort == "ascending":
            sort = -1 #ascending means from latest to earliest run
        else:
            sort = 1

        col = self.db.aggregate(
                    [
                        {"$match": {'$and': [{'start': {'$gte':ts_beg}}, {'start': {'$lte':ts_end}}]}},
                        {"$unwind": "$data"},
                        {"$project": {
                                        "name": 1, "number": 1, "host": "$data.host",
                                    }},
                        {"$match": {"host": host}},
                        {"$project": {"name":1, "number":1}},
                        {"$sort": {"start": sort}}
                    ]
                )

        return col

    def GetClearance(self, ts_beg=None, ts_end=None, sort="ascending"):
        """Function: GetClearance(...)

        Returns from a mongoDB database all entries which have a list of dictionaries (data field) and an individual
        data dictionary contains a key "status" with the value "RucioClearance"

        :param ts_beg: A datetime object to set a time interval - Start value (field 'start')
        :param ts_end: A datetime object to set a time interval - Stop value (field 'start')
        :param sort: Sort your mongodb aggregation ascending (standard) or descending
        :return collection: The summary collection of the requested status (RucioClearance) with name and number
                            field only. Acts as reduced summary
        """
        if ts_beg == None:
            ts_beg = self.GetSmallest("start")
        if ts_end == None:
            ts_end = self.GetLargest("start")

        #evaluate sort direction:
        if sort == "ascending":
            sort = -1 #ascending means from latest to earliest run
        else:
            sort = 1

        collection = self.db.aggregate(
                    [
                        {"$match": {'$and': [ {'start': {'$gte':ts_beg}}, {'start': {'$lte':ts_end}} ]}},
                        {"$unwind": "$data"},
                        {"$project": {"name":1, "number":1, "status": "$data.status"}},
                        {"$match": {"status": "RucioClearance"}},
                        {"$project": {"name":1, "number":1, "data":1}},
                        {"$sort": {"start": sort}}
                    ]
                )

        return collection

    def StatusDataField(self, id_field, type=None, host=None):
        #Test if a dictionary exists in a list for a specific combination of type and host
        run = self.GetRunByID(id_field)[0]

        value = False
        for i_run in run['data']:
            if i_run['type'] != type:
                continue
            if i_run['host'] != host:
                continue
            value=True

        return value

    def ShowDataField(self, id_field, type=None, host=None):
        #Test if a dictionary exists in a list for a specific combination of type and host
        run = self.GetRunByID(id_field)[0]

        for i_run in run['data']:
            if i_run['type'] != type:
                continue
            if i_run['host'] != host:
                continue
            print(" [-> ", i_run['type'], "/", i_run['host'], " <-]")
            if 'rse' in i_run:
                print(" [-> ", i_run['rse'])
            else:
                print(" [->  No RSE information")
            if 'status' in i_run:
                print(" [-> ", i_run['status'])
            else:
                print(" [->  No status information")
            if 'meta' in i_run:
                print(" [-> ", i_run['meta'])
            else:
                print(" [->  No meta information")

    def SetDataField(self, id_field, type=None, host=None,
                     key=None, value=None, new=False):
        """
        Updates data entries for the given run to have `key` == `value`, where the `type`
        and `host` fields match. If `new` is False, only entries with the key are changed

        :param id_field: the ObjectId of the run to update
        :param type: str, the `type` field of entries to update
        :param host: str, the `host` field of entries to update
        :param key: str, the key to be changed or added
        :param value: the value to be set
        :param new: bool, if False, only changes entries where the key exists
        :return None:
        """

        if type != None and host!=None and key != None:
            array_filters = [{"elem.type" : type, "elem.host" : host}]
            if not new:
                array_filters[0]["elem.{}".format(key)] =  {"$exists" : 1}
            self.db.update_one({"_id" : id_field},
                    {"$set" : {"data.$[elem].{}".format(key) : value}},
                    {"arrayFilters" : array_filters})

    def AddDatafield(self, id_field, new_dict):
        """
        Adds the specified new entry into the run with the given id

        :param id_field: the ObjectId of the run to update
        :param new_dict: the new data entry to add, min format {type : str, host : str, location : str}
        :return None:
        """

        self.db.update_one({"_id": id_field},
                              {"$push": {"data": new_dict}})

    def RemoveDatafield(self, id_field, rem_dict):
        """
        Moves the data entry specified from the "active" list to the deleted list.
        Entries are updated with the time and entity responsible for the removal

        :param id_field: the ObjectId of the run to update
        :param rem_dict: the data entry to remove, min format {type : str, host : str, location : str}
        :return None:
        """

        new_entry = dict(rem_dict.items())
        new_entry.update({"at" : datetime.datetime.utcnow(), "by" : "admix"})

        self.db.update_one({"_id" : id_field},
                              {"$pull" : {"data" : rem_dict},
                               "$push" : {"deleted_data" : new_entry}})

    def UpdateDatafield(self, id_field, new_data=None):
        run = self.GetRunByID(id_field)[0]

        old_data = run['data']

        if old_data==new_data:
            return 0
        else:
            list_diff = [x for x in new_data if x not in old_data]

            for i in list_diff:
                print("New destination for host {host}:".format(host=i['host']))
                print(" -> {dest}".format(dest=i['destination']))
            self.db.update_one({"_id": id_field},
                                    {"$set": {"data": new_data}})


    def WriteDestination(self, id_field, type=None, host=None, destination=None):
        run = self.GetRunByID(id_field)[0]

        new_data = run['data']

        for i_run in new_data:
            if 'destination' not in i_run:
                i_run['destination']= []

        if type != None and host!=None and destination != None:
            for i_run in new_data:
                if i_run['type'] != type:
                    continue
                if i_run['host'] != host:
                    continue
                if destination in i_run['destination']:
                    break
                print("Change a destination for:", type, "and", host, "to", destination)
                i_run['destination'].append(destination)

            print("The new destination is")
            for i_run in new_data:
                print(" <> ", i_run['host'], i_run['location'], i_run['destination'])


            self.db.find_one_and_update({"_id": id_field},
                                        {"$set": {"data": new_data}})

            print("after new set (test)")
            #run = self.GetRunByID(id_field)[0]
            #for i_run in run['data']:
                #print(" <new> -", i_run['host'], i_run['location'], i_run['destination'])

    def SetStatus(self, number, status):
        self.db.update_one({'number': number},
                                  {'$set': {'status': status}}
                                  )

    def GetHashByContext(self, context, dtype, reconnect=False):

        if reconnect == True:
            self.Connect()

        query = {"name": context}
        selected_context = list(self.contexts.find(query).sort([("date_added", -1)]).limit(1))[0]
        return selected_context["hashes"][dtype]




# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater
from admix.utils.naming import make_did

def list_file_replicas(run_number, dtype, hash, rse='UC_DALI_USERDISK'):

    db = ConnectMongoDB()
    rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))

#    print("Looking for run "+str(run_number)+", data type "+dtype+", hash "+hash+", in rse="+rse)

    # checks if run is present in run database
    # this will improve the reaction speed in case the run is not existing
    # since we do not call Rucio commands
    cursor = db.GetRunByNumber(run_number)        
    if len(cursor)==0:
#        print("Error. Run not existing in database")
        return list()

    # build did
    did = make_did(run_number, dtype, hash)

    file_replicas = {}

    # check if the did esists in the given rse
    if rc.CheckRule(did, rse) != 'OK':
#        print("Error. Not found in this rse")
        return list()

    file_replicas = rc.ListFileReplicas(did,rse,localpath=True)

    return list(file_replicas.values())




# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper
import time
import shutil

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.utils import make_did
from admix.utils.list_file_replicas import list_file_replicas
import pymongo

@Collector
class Upload():

    def __init__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')

        open("/tmp/admix-upload_from_lngs", 'a').close()

        #Take all data types categories
        self.NORECORDS_DTYPES = helper.get_hostconfig()['norecords_types']
        self.RAW_RECORDS_DTYPES = helper.get_hostconfig()['raw_records_types']
        self.RECORDS_DTYPES = helper.get_hostconfig()['records_types']

        # Choose which RSE you want upload to
        self.UPLOAD_TO = helper.get_hostconfig()['upload_to']

        #Choose which data type you want to treat
        self.DTYPES = self.NORECORDS_DTYPES + self.RECORDS_DTYPES + self.RAW_RECORDS_DTYPES

        if helper.global_dictionary.get('high'):
               self.DTYPES = self.NORECORDS_DTYPES

        if helper.global_dictionary.get('low'):
               self.DTYPES = self.RECORDS_DTYPES + self.RAW_RECORDS_DTYPES


        self.DATADIR = helper.get_hostconfig()['path_data_to_upload']
        self.periodic_check = helper.get_hostconfig()['upload_periodic_check']

        #Init the runDB
        self.db = ConnectMongoDB()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()
        self.rc.SetProxyTicket("rucio_x509")


    def find_next_run_to_upload(self):
#        cursor = self.db.db.find({'status': 'eb_ready_to_upload', 'bootstrax.state': 'done' }, {'number': 1, 'data': 1})
        cursor = self.db.db.find({'status': { '$in': ['eb_ready_to_upload','transferring']}, 'bootstrax.state': 'done' }, {'number': 1, 'data': 1})
        id_run = 0
        min_run = float('inf')

        for run in cursor:
            print(run['number'])
            if run['number']<10000:
                continue
#            if run['number'] not in [10161]:
#                 continue
            if run['number'] < min_run:
                min_run = run['number']
                id_run = run['_id']
        print("   ",min_run)
        return id_run



    def find_next_run_and_dtype_to_upload(self):
        cursor = self.db.db.find({'status': { '$in': ['eb_ready_to_upload','transferring']}, 'bootstrax.state': 'done' }, {'number': 1, 'data': 1, 'bootstrax': 1}).sort('number',pymongo.ASCENDING)
        id_run = 0
        min_run = float('inf')

        cursor = list(cursor)

        helper.global_dictionary['logger'].Info('\t==> Runs in queue: {0}'.format(len(cursor)))

        for run in cursor:

            # Get run number
            number = run['number']

            # Forget about old runs
            if number<10000:
                continue

            # For debugging: select a specific run
#            if number not in [10161]:
#                continue

            # Extracts the correct Event Builder machine who processed this run
            bootstrax = run['bootstrax']
            eb = bootstrax['host'].split('.')[0]

            # Look for the first data type available to be uploaded
            datum = None
            for dtype in self.DTYPES:

                #if dtype != "raw_records_aqmon":
                #    continue

                # search if dtype still has to be uploaded
                for d in run['data']:
                    if d['type'] == dtype and eb in d['host'] and ('status' not in d or ('status' in d and d['status'] == 'eb_ready_to_upload')):
                        datum = d
                        continue
 
                if datum is not None:
                    continue

            # If there is a candidate data type, return run_id and data type
            if datum is not None:
                return run['_id'], datum

        return 0,''




    def add_rule(self,run_number, dtype, hash, rse, datum=None, lifetime=None, update_db=True):
        did = make_did(run_number, dtype, hash)
        if dtype in self.NORECORDS_DTYPES:
            priority = 1
        else:
            priority = 3
        result = self.rc.AddRule(did, rse, lifetime=lifetime, priority=priority)
        #if result == 1:
        #   return
        helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}: rule added: {2} ---> {3}'.format(run_number,dtype,did,rse))

        if update_db:
            self.db.db.find_one_and_update({'number': run_number},
                                      {'$set': {'status': 'transferring'}}
                                  )

            rucio_rule = self.rc.GetRule(did, rse=rse)
            updated_fields = {'host': "rucio-catalogue",
                         'type': dtype,
                         'location': rse,
                         'lifetime': rucio_rule['expires'],
                         'status': 'transferring',
                         'did': did,
                         'protocol': 'rucio'
            }

            if datum == None:
                data_dict = updated_fields
            else:
                data_dict = datum.copy()
                data_dict.update(updated_fields)

            docid = self.db.db.find_one({'number': run_number}, {'_id': 1})['_id']
            self.db.AddDatafield(docid, data_dict)


    def add_conditional_rule(self,run_number, dtype, hash, from_rse, to_rse, datum=None, lifetime=None, update_db=True):
        did = make_did(run_number, dtype, hash)
        if dtype in self.NORECORDS_DTYPES:
            priority = 1
        else:
            priority = 3
        result = self.rc.AddConditionalRule(did, from_rse, to_rse, lifetime=lifetime, priority=priority)
        #if result == 1:
        #   return
        helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}: conditional rule added: {2} ---> {3}'.format(run_number,dtype,did,to_rse))

        if update_db:
            self.db.db.find_one_and_update({'number': run_number},
                                      {'$set': {'status': 'transferring'}}
                                  )

            rucio_rule = self.rc.GetRule(did, rse=to_rse)
            updated_fields = {'host': "rucio-catalogue",
                         'type': dtype,
                         'location': to_rse,
                         'lifetime': rucio_rule['expires'],
                         'status': 'transferring',
                         'did': did,
                         'protocol': 'rucio'
            }

            if datum == None:
                data_dict = updated_fields
            else:
                data_dict = datum.copy()
                data_dict.update(updated_fields)

            docid = self.db.db.find_one({'number': run_number}, {'_id': 1})['_id']
            self.db.AddDatafield(docid, data_dict)



    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')

        # Get a new run to upload
        id_to_upload, datum = self.find_next_run_and_dtype_to_upload()
        if id_to_upload == 0:
            helper.global_dictionary['logger'].Info('\t==> No data type available to upload')
            return 0

        # Load the run
        run = self.db.db.find_one({'_id': id_to_upload}, {'number': 1, 'data': 1, 'bootstrax': 1})

        # Get run number
        number = run['number']

        # Set run status to "transferring"
        self.db.SetStatus(number, 'transferring')

        # Extracts the correct Event Builder machine who processed this run
        bootstrax = run['bootstrax']
        eb = bootstrax['host'].split('.')[0]

#        # Performs upload on selected run
#        helper.global_dictionary['logger'].Info('Uploading run {0} from {1}'.format(number,eb))


        # Attempting to book this data type
        dtype = datum['type']
        helper.global_dictionary['logger'].Info('\t==> Run {0}, trying to book data type {1} for uploading. Starting match'.format(number,dtype))

        # Books the eb data entry by setting its status to the PID of the process
        PID = str(os.getpid())
        self.db.db.find_one_and_update({'_id': id_to_upload,'data': {'$elemMatch': datum}},
                                       {'$set': {'data.$.status': PID}})

        # Wait for 10 seconds
        time.sleep(10)

        # Then check if this status is still equal to the same PID
        run = self.db.db.find_one({'_id': id_to_upload}, {'number': 1, 'data': 1, 'bootstrax': 1})
        datum = None
        for d in run['data']:
            if d['type'] == dtype and eb in d['host'] and d['status']==PID:
                datum = d
                continue
        
        # If there is no data available to upload any more, exit
        if datum is None:
           helper.global_dictionary['logger'].Info('\t==> Run {0}, lost challenge, data type {1} not available any more'.format(number,dtype))
           return 0

        # Match won. Proceeding with uploading
        helper.global_dictionary['logger'].Info('\t==> Run {0}, match won, starting uploading data type {1}'.format(number,dtype))

        # Modify data type status to "transferring"
        self.db.db.find_one_and_update({'_id': id_to_upload,'data': {'$elemMatch': datum}},
                                       {'$set': {'data.$.status': "transferring"}})

        # Wait for 3 seconds
        time.sleep(3)

        # Reloading the updated datum
        run = self.db.db.find_one({'_id': id_to_upload}, {'number': 1, 'data': 1, 'bootstrax': 1})
        datum = None
        for d in run['data']:
            if d['type'] == dtype and eb in d['host']:
                datum = d
                continue

        # Check, for coherency, if there is no rucio entry in DB for this data type
        in_rucio_upload_rse = False
        in_rucio_somewhere_else = False
        for d in run['data']:
            if d['type'] == datum['type'] and d['host'] == 'rucio-catalogue' and d['location'] == self.UPLOAD_TO:
                in_rucio_upload_rse = True
            if d['type'] == datum['type'] and d['host'] == 'rucio-catalogue' and d['location'] != self.UPLOAD_TO:
                in_rucio_somewhere_else = True

        if in_rucio_upload_rse:
            helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1} has already a DB entry for RSE {2}. Forced to stop'.format(number,dtype,self.UPLOAD_TO))
            return 0

        if in_rucio_somewhere_else:
            helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1} has already a DB entry for some external RSE. Forced to stop'.format(number,dtype))
            return 0

        # Preparing relevant info for uploading
        
        file = datum['location'].split('/')[-1]
        hash = file.split('-')[-1]
        upload_path = os.path.join(self.DATADIR, eb, file)
        did = make_did(number, dtype, hash)

        # Querying Rucio: if a rule exists already for this DID on LNGS, skip uploading
        rucio_rule = self.rc.GetRule(upload_structure=did, rse=self.UPLOAD_TO)
        if rucio_rule['exists']:
            helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1} has already a Rucio rule for RSE {2}. Forced to stop'.format(number,dtype,self.UPLOAD_TO))
            return 0
            
        # Finally, start uploading with Rucio
        helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}, start uploading, DID = {2}'.format(number,dtype,did))
        result = self.rc.Upload(did, upload_path, self.UPLOAD_TO, lifetime=None)
        helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}, uploaded, DID = {2}'.format(number,dtype,did))

        # Wait for 10 seconds
        time.sleep(10)
  
        # Checking the status of this new upload rule
        rucio_rule = self.rc.GetRule(upload_structure=did, rse=self.UPLOAD_TO)
        if rucio_rule['state'] != 'OK':
            helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}, according to Rucio, uploading failed. Forced to stop'.format(number,dtype))
            exit()

        # Checking the actual number of files uploaded
#        nfiles = len(list_file_replicas(number, dtype, hash, rucio_rule['rse']))
#        if nfiles != datum['file_count']:
#            helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}, unconsistent number of files (Rucio: {2}, DB: {3}). Forced to stop'.format(number,dtype,nfiles,datum['file_count']))
#            exit()        

        # Update the eb data entry with status "transferred"
        self.db.db.find_one_and_update({'_id': id_to_upload,'data': {'$elemMatch': datum}},
                                       {'$set': {'data.$.status': "transferred"}})

        # Add a new data field with LNGS as RSE and with status "trasferred"
        data_dict = datum.copy()
        data_dict.update({'host': "rucio-catalogue",
                          'type': dtype,
                          'location': self.UPLOAD_TO,
                          'lifetime': rucio_rule['expires'],
                          'status': 'transferred',
                          'did': did,
                          'protocol': 'rucio'
                      })
        self.db.AddDatafield(run['_id'], data_dict)

        # set a rule to ship data on GRID
        if rucio_rule['state'] == 'OK':
            if dtype in self.NORECORDS_DTYPES:
                self.add_rule(number, dtype, hash, 'UC_DALI_USERDISK',datum=datum)
                self.add_conditional_rule(number, dtype, hash, 'UC_DALI_USERDISK', 'UC_OSG_USERDISK',datum=datum)
            else:
                self.add_rule(number, dtype, hash, 'UC_OSG_USERDISK',datum=datum)
#                if dtype in self.RECORDS_DTYPES:
#                    self.add_conditional_rule(number, dtype, hash, 'UC_OSG_USERDISK', 'UC_DALI_USERDISK',datum=datum)
                self.add_conditional_rule(number, dtype, hash, 'UC_OSG_USERDISK', 'CCIN2P3_USERDISK',datum=datum)
                if dtype in self.RAW_RECORDS_DTYPES:
                    self.add_conditional_rule(number, dtype, hash, 'UC_OSG_USERDISK', 'CNAF_TAPE2_USERDISK',datum=datum)


        return 0


    def __del__(self):
        pass



# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater

import copy

@Collector
class UpdateRunDBMongoDB():

    def __init__(self):
        pass

    def __del__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')

        self.db = ConnectMongoDB()
        self.db.Connect()

        # We want the first and the last run:
        self.gboundary = self.db.GetBoundary()
        self.run_nb_min = self.gboundary['min_number']
        self.run_nb_max = self.gboundary['max_number']
        self.run_ts_min = self.gboundary['min_start_time']
        self.run_ts_max = self.gboundary['max_start_time']

        #Init the Rucio data format evaluator in three steps:
        self.rc_reader = ConfigRucioDataFormat()
        self.rc_reader.Config(helper.get_hostconfig('rucio_template'))
        self.defined_types = self.rc_reader.GetTypes()

        #This class will evaluate your destinations:
        self.destination = Destination()

        #Since we deal with an experiment, everything is predefine:
        self.exp_temp = Templater()
        self.exp_temp.Config(helper.get_hostconfig()['template'])

        #Init a class to handle keyword strings:
        self.keyw = Keyword()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()



    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')

        ts_beg = None
        ts_end = None
        if helper.global_dictionary.get('run_numbers') != None:
            #Evaluate terminal input for run number assumption (terminal input == true)
            true_nb_beg, true_nb_end = helper.eval_run_numbers(helper.global_dictionary.get('run_numbers'),
                                                               self.run_nb_min,
                                                               self.run_nb_max)
            #Get the timestamps from the run numbers:
            ts_beg = self.db.FindTimeStamp('number', int(true_nb_beg))
            ts_end = self.db.FindTimeStamp('number', int(true_nb_end))

        elif helper.global_dictionary.get('run_timestamps') != None:
            #Evaluate terminal input for run name assumption
            true_ts_beg, true_ts_end = helper.eval_run_timestamps(helper.global_dictionary.get('run_timestamps'),
                                                               self.run_ts_min,
                                                               self.run_ts_max)
            ts_beg = true_ts_beg
            ts_end = true_ts_end

        elif helper.global_dictionary.get('run_timestamps') == None and \
            helper.global_dictionary.get('run_numbers') == None:
            ts_beg = self.run_ts_min
            ts_end = self.run_ts_max
        else:
            helper.global_dictionary['logger'].Error("Check for your input arguments (--select-run-number or --select-run-time")
            exit(1)
            #exection

        #Get your collection of run numbers and run names
        collection = self.db.GetRunsByTimestamp(ts_beg, ts_end)



        # We need to define the right plugins here which are getting updated:
        # Therefore we have two options!
        # 1) Use all types which are defined in Rucio format configuration file
        # 2) Use the command line argument inputs
        if helper.global_dictionary.get('type') != None and isinstance(helper.global_dictionary.get('type'), list):
            plugin_list = helper.global_dictionary.get('type')
        else:
            plugin_list = self.defined_types

        #plugin_list = ['raw_records', 'records']

        #Run through the overview collection:
        for i_run in collection:
            #Extract run number and name from overview collection
            r_name = i_run['name']

            dict_name = {}
            dict_name['date'] = r_name.split("_")[0]
            dict_name['time'] = r_name.split("_")[1]

            r_number = i_run['number']
            #Pull the full run information (according to projection which is pre-defined) by the run name
            db_info = self.db.GetRunByName(r_name)[0]

            if 'data' not in db_info:
                continue

            helper.global_dictionary['logger'].Info(f"Update data set {r_name}/{r_number}")
            #adjust the data field with a Rucio update if necessary:
            for i_data in db_info['data']:

                if i_data['host'] != 'rucio-catalogue':
                    continue

                if i_data['type'] not in plugin_list:
                    continue

                # check rules
                if 'location' not in i_data or i_data['location'] == "n/a":
                    helper.global_dictionary['logger'].Info("Skip {0} - Check for location and data consistency".format(i_data['type']))
                    continue

                # Go for the rucio template:
                #ToDo We should do better than deepcopy
                rucio_template = copy.deepcopy(self.rc_reader.GetPlugin(i_data['type']))

                # Fill the key word class with information beforehand:
                #extract the hash
                try:
                    _hash = i_data['location'].split(":")[1].split("-")[1]
                except:
                    _hash = "Xenon1T"

                #This section is adjust for XENON1T legacy support:
                # The detector name muon_veto from the database changes into mv
                # Fix one minute bug from past XENON1T days...
                # Fix the science run definition hole between SR0 and SR1
                sr_run_selected = helper.get_science_run(db_info['start'])
                if _hash == 'Xenon1T':
                    db_info = helper.xenon1t_detector_renamer(db_info)
                    #fixing the one minute offset bug in XENON1T by overwriting the dict_name from true
                    #Rucio locations such they are stored in the runDB
                    dict_name['date'] = i_data['location'].split(":")[0].split("_")[2]
                    dict_name['time'] = i_data['location'].split(":")[0].split("_")[3]
                    sr_run_selected = i_data['location'].split(":")[0].split("_")[1].replace("SR", "")


                #Fill in the templates:
                self.keyw.ResetTemplate()
                self.keyw.SetTemplate(dict_name)
                self.keyw.SetTemplate(db_info)
                self.keyw.SetTemplate({'hash': _hash, 'plugin': i_data['type']})
                self.keyw.SetTemplate({'science_run': sr_run_selected})

                rucio_template = self.keyw.CompleteTemplate(rucio_template)
                rc_rules = self.rc.ListDidRules(rucio_template)

                rule = []
                for i_rule in rc_rules:
                    #Extract the rucio rule information per RSE location
                    #and append

                    rse_expression = i_rule['rse_expression']
                    state = i_rule['state']
                    #print(i_rule)
                    if i_rule['expires_at'] != None:
                        expires_at = i_rule['expires_at'].strftime("%Y-%m-%d-%H:%M:%S")
                    else:
                        expires_at = None
                    rule.append("{rse}:{state}:{lifetime}".format(rse=rse_expression,
                                                                      state=state,
                                                                      lifetime=expires_at))


                #Update RSE field if necessary:
                if rule != i_data['rse'] and len(rule) >= 1:
                    self.db.SetDataField(db_info['_id'], type=i_data['type'], host=i_data['host'], key='rse', value=rule)
                    helper.global_dictionary['logger'].Info("Updated location ({0}) for type {1} to {2}".format(i_data['location'], i_data['type'], rule))

                    #Since we update the runDB with new Rucio locations, I suppose we should update the status field
                    #just in case...
                    if self.db.GetDataField(db_info['_id'], type=i_data['type'], host=i_data['host'],
                                            key='status') != "transferred":
                        self.db.SetDataField(db_info['_id'], type=i_data['type'], host=i_data['host'], key='status',
                                             value="transferred")
                elif rule != i_data['rse'] and len(rule) == 0:
                    helper.global_dictionary['logger'].Info("No location ({0}) for type {1} to {2} found in Rucio".format(i_data['location'], i_data['type'], rule))
                    helper.global_dictionary['logger'].Info("Set for purge: status -> NoRucioEntry")
                    self.db.SetDataField(db_info['_id'], type=i_data['type'], host=i_data['host'], key='status',
                                         value="RucioClearance")
                    self.db.SetDataField(db_info['_id'], type=i_data['type'], host=i_data['host'], key='rse', value=rule)

                else:
                    helper.global_dictionary['logger'].Info("No update at location ({0}) for type {1} to {2}".format(i_data['location'], i_data['type'], rule))



# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper
import time
import shutil

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.utils import make_did
from admix.utils.list_file_replicas import list_file_replicas

@Collector
class MonitorRun():

    def __init__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')


        #Define data types
        self.NORECORDS_DTYPES = helper.get_hostconfig()['norecords_types']
        self.RAW_RECORDS_DTYPES = helper.get_hostconfig()['raw_records_types']
        self.RECORDS_DTYPES = helper.get_hostconfig()['records_types']

        #Get other parameters
        self.DATADIR = helper.get_hostconfig()['path_data_to_upload']
        self.periodic_check = helper.get_hostconfig()['upload_periodic_check']
        self.RSES = helper.get_hostconfig()['rses']

        self.minimum_number_acceptable_rses = 2

        # Storing some backup hashes in case DID information is not available
        self.bkp_hashes = { 'raw_records':'rfzvpzj4mf', 'raw_records_he':'rfzvpzj4mf', 'raw_records_mv':'rfzvpzj4mf', 'raw_records_aqmon':'rfzvpzj4mf', 'records':'56ausr64s7' }


        #Init the runDB
        self.db = ConnectMongoDB()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()
        self.rc.SetProxyTicket("rucio_x509")




    def run(self,*args, **kwargs):
        print('Run task '+self.__class__.__name__)

        data_types = self.RAW_RECORDS_DTYPES + self.RECORDS_DTYPES + self.NORECORDS_DTYPES
#        data_types = self.RAW_RECORDS_DTYPES + self.RECORDS_DTYPES


        # Get all runs that are already transferred and that still have some data_types in eb 
        cursor = self.db.db.find({
#            'number': 7448
#            'number': 8322
#            'number': 6371
#            'number': 7148
#            'number': 8590
#            'number': 8626
            'number': 7148
#            'number': {'$gte': 8100, '$lt': 8200}
        },
        {'_id': 1, 'number': 1, 'data': 1, 'status':1, 'bootstrax': 1})

        cursor = list(cursor)

        print('Runs that will be processed are {0}'.format([c["number"] for c in cursor]))

        # Runs over all listed runs
        for run in cursor:

            # Gets run number
            number = run['number']
            print('Run: {0}'.format(number))

            # Extracts the correct Event Builder machine who processed this run                                                                               
            bootstrax = run['bootstrax']
            eb = bootstrax['host'].split('.')[0]
            print('Processed by: {0}'.format(eb))

            # Get the status
            if 'status' in run:
                print('Status: {0}'.format(run['status']))
            else:
                print('Status: {0}'.format('Not available'))

            # Merges data and deleted_data
            if 'deleted_data' in run:
                data = run['data'] + run['deleted_data']
            else:
                data = run['data']
            
            # Check is there are more instances in more EventBuilders
            extra_ebs = set()
            for d in data:
                if 'eb' in d['host'] and eb not in d['host']: 
                    extra_ebs.add(d['host'].split('.')[0])
            if len(extra_ebs)>0:
                print('\t\t Warning : The run has been processed by more than one EventBuilder: {0}'.format(extra_ebs))

            # Runs over all data types to be monitored
            for dtype in data_types:

                # Data type name
                print('{0}'.format(dtype))

                # Take the official number of files accordingto run DB
                Nfiles = -1
                for d in data:
                    if d['type'] == dtype and eb in d['host']:
                        if 'file_count' in d:
                            Nfiles = d['file_count']
                if Nfiles == -1:
                    print('\t Number of files: missing in DB')
                else:
                    print('\t Number of files: {0}'.format(Nfiles))

                # Check if data are still in the data list and not in deleted_data
                DB_InEB = False
                for d in run['data']:
                    if d['type'] == dtype and eb in d['host']:
                        DB_InEB = True
                DB_NotInEB = False
                if 'deleted_data' in run:
                    for d in run['deleted_data']:
                        if d['type'] == dtype and eb in d['host']:
                            DB_NotInEB = True
                if DB_InEB and not DB_NotInEB:
                    print('\t DB : still in EB')
                if not DB_InEB and DB_NotInEB:
                    print('\t DB : deleted from EB')
                if DB_InEB and DB_NotInEB:
                    print('\t\t Incoherency in DB: it is both in data list and in deleted_data list')
#                if (DB_InEB and DB_NotInEB) or (not DB_InEB and not DB_NotInEB):
#                    print('\t\t incoherency in DB: it is neither in data list nor in deleted_data list')

                # Check if data are still in the EB disks without using the DB
                upload_path = ""
                for d in run['data']:
                    if d['type'] == dtype and eb in d['host']:
                        file = d['location'].split('/')[-1]
                        upload_path = os.path.join(self.DATADIR, eb, file) 
                path_exists = os.path.exists(upload_path)
                if upload_path != "" and path_exists:
                    path, dirs, files = next(os.walk(upload_path))
                    print('\t Disk: still in EB disk and with',len(files),'files')
                else:
                    print('\t Disk: not in EB disk anymore')
                if DB_InEB and not path_exists:
                    print('\t\t Incoherency in DB and disk: it is in DB data list but it is not in the disk')
                if DB_NotInEB and path_exists:
                    print('\t\t Incoherency in DB and disk: it is in DB deleted_data list but it is still in the disk')

                # The list of DIDs (usually just one)
                dids = set()
                for d in data:
                    if d['type'] == dtype and d['host'] == 'rucio-catalogue':
                        if 'did' in d:
                            dids.add(d['did'])
                print('\t DID:', dids)

                # Check the presence in each available RSE
                Nrses = 0
                for rse in self.RSES:
                    is_in_rse = False
                    for d in run['data']:
                        if d['type'] == dtype and rse in d['location']:
                            if 'status' in d:
                                status = d['status']
                            else:
                                status = 'Not available'
                            if 'did' in d:
                                hash = d['did'].split('-')[-1]
                                did = d['did']
                            else:
                                print('\t\t Warning : DID information is absent in DB data list (old admix version). Using standard hashes for RSEs')
                                hash = self.bkp_hashes.get(dtype)
                                did = make_did(number, dtype, hash)
                            rucio_rule = self.rc.GetRule(upload_structure=did, rse=rse)
                            files = list_file_replicas(number, dtype, hash, rse)
                            if rucio_rule['exists']:
                                print('\t', rse+': DB Yes, Status',status,', Rucio Yes, State',rucio_rule['state'],",",len(files), 'files')
                                if len(files) < Nfiles:
                                    print('\t\t Warning : Wrong number of files in Rucio!!!')
                            else:
                                print('\t', rse+': DB Yes, Status',status,', Rucio No')
                            # print(files)
                            is_in_rse = True
                            Nrses += 1
                    if not is_in_rse:
#                        print('\t\t Warning : data information is absent in DB data list. Trying using standard hashes to query Rucio')
                        hash = self.bkp_hashes.get(dtype)
                        did = make_did(number, dtype, hash)
                        rucio_rule = self.rc.GetRule(upload_structure=did, rse=rse)
                        files = list_file_replicas(number, dtype, hash, rse)
                        if rucio_rule['exists']:
                            print('\t', rse+': DB No, Rucio Yes, State',rucio_rule['state'],",",len(files), 'files')
                            if len(files) < Nfiles:
                                print('\t\t Warning : Wrong number of files in Rucio!!!')
                        else:
                            print('\t', rse+': DB No, Rucio No')
                print('\t Number of sites: ', Nrses)
                    


        return 0


    def __del__(self):
        pass



# -*- coding: utf-8 -*-
import json
import os
import shutil
from admix.helper import helper

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

# get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater


@Collector
class PurgeMongoDB():
    """Class: PurgeMongoDB
    Purpose of this class is to purge physical data locations outside of Rucio which are registered to
    a meta database.

    Due to the complex file structure our data products the purge command became quite extensive in terms of
    selections to narrow down the datasets (depending on type, version (hash), host and location).

    PurgeMongoDB runs a purging process only one type of data (e.g. raw_records) at the same time. Multiple types
    are not supported (yet?).

    Therefore you can run this command with these inputs:

    admix PurgeMongoDB --admix-config /path/to/config/file.config
                       --select-run-times <date>_<time> (as single, multi or range)
                          or
                       --select-run-number XXXXX (as single or range)
                       --type data_type (e.g. raw_records)
                       --hash XXXXXXXXXX (10 character hash sum which represents a certain version of the data type
                       --force (Allows you enforce manual purge mode)
    Full example:
    admix PurgeMongoDB --once --admix-config  /home/bauermeister/Development/software/admix_config/host_config_dali_olddb.config --select-run-times 171225_0040-171225_0140 --type raw_records records --hash 7k65yaooed --force

    """
    def __init__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')

        # Init the runDB
        self.db = ConnectMongoDB()
        self.db.Connect()

        # We want the first and the last run:
        self.gboundary = self.db.GetBoundary()
        self.run_nb_min = self.gboundary['min_number']
        self.run_nb_max = self.gboundary['max_number']
        self.run_ts_min = self.gboundary['min_start_time']
        self.run_ts_max = self.gboundary['max_start_time']

        # Init the Rucio data format evaluator in three steps:
        self.rc_reader = ConfigRucioDataFormat()
        self.rc_reader.Config(helper.get_hostconfig('rucio_template'))

        # This class will evaluate your destinations:
        self.destination = Destination()

        # Since we deal with an experiment, everything is predefine:
        self.exp_temp = Templater()
        self.exp_temp.Config(helper.get_hostconfig()['template'])

        # Init a class to handle keyword strings:
        self.keyw = Keyword()

        # Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()

    def run(self, *args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')

        # Check for type pre-selection
        d_type = None
        d_hash = None
        if helper.global_dictionary.get("type") is None:
            helper.global_dictionary['logger'].Info(
                'Selected type of data to purge: {0}'.format(helper.global_dictionary.get("type")))
            helper.global_dictionary['logger'].Info("No data type is selected! Nothing to purge -> Finish here")
            return 1
        else:
            d_type = helper.global_dictionary.get("type")[0]

        if helper.global_dictionary.get("hash") is None:
            helper.global_dictionary['logger'].Info(
                'Selected type of data to purge: {0}'.format(helper.global_dictionary.get("hash")))
            helper.global_dictionary['logger'].Info(
                "No data hash (version) is selected! Nothing to purge -> Finish here")
            return 1
        else:
            d_hash = helper.global_dictionary.get("hash")[0]

        ts_beg = None
        ts_end = None
        if helper.global_dictionary.get('run_numbers') != None:
            # Evaluate terminal input for run number assumption (terminal input == true)
            true_nb_beg, true_nb_end = helper.eval_run_numbers(helper.global_dictionary.get('run_numbers'),
                                                               self.run_nb_min,
                                                               self.run_nb_max)
            # Get the timestamps from the run numbers:
            ts_beg = self.db.FindTimeStamp('number', int(true_nb_beg))
            ts_end = self.db.FindTimeStamp('number', int(true_nb_end))

        elif helper.global_dictionary.get('run_timestamps') != None:
            # Evaluate terminal input for run name assumption
            true_ts_beg, true_ts_end = helper.eval_run_timestamps(helper.global_dictionary.get('run_timestamps'),
                                                                  self.run_ts_min,
                                                                  self.run_ts_max)
            ts_beg = true_ts_beg
            ts_end = true_ts_end

        elif helper.global_dictionary.get('run_timestamps') == None and \
            helper.global_dictionary.get('run_numbers') == None:
            ts_beg = self.run_ts_min
            ts_end = self.run_ts_max
        else:
            helper.global_dictionary['logger'].Error(
                "Check for your input arguments (--select-run-number or --select-run-time")
            exit(1)
            # exection

        # After we know the times:
        helper.global_dictionary['logger'].Info(f"Run between {ts_beg} and {ts_end}")

        print("Purge Module")
        host_name = helper.get_hostname()
        hc_host = helper.get_hostconfig("host")  # Supposed to be the database entry to identfy locations

        # Get your collection of run numbers and run names what contain a certain host of choice
        # hint: You are allowed to get several plugins (types) and versions (hashsums) for the same
        #       dataset at a certain host.
        collection = self.db.GetHosts(hc_host, ts_beg, ts_end)

        # Run through the overview collection:
        for i_run in collection:

            # Extract run number and name from overview collection
            r_name = i_run['name']
            r_number = i_run['number']

            # Pull the full run information (according to projection which is pre-defined) by the run name
            db_info = self.db.GetRunByName(r_name)[0]

            print(db_info['name'], db_info['number'])

            # Data are supposed to have enough rucio copies:
            valid_rucio_copies = 0
            valid_rucio_rses = ""
            for i_data in db_info['data']:
                if i_data['host'] != "rucio-catalogue":
                    continue
                if i_data['type'] != d_type:
                    continue
                if d_hash not in i_data['location']:
                    continue
                if i_data['status'] != 'transferred':
                    continue
                valid_rucio_copies = sum([1 if i_rse.split(":")[1] == 'OK' else 0 for i_rse in i_data['rse']])
                valid_rucio_rses = [i_rse.split(":")[0] if i_rse.split(":")[1] == 'OK' else '' for i_rse in i_data['rse']]
            valid_rucio_rses = ', '.join(list(filter(None, valid_rucio_rses)))


            valid_disk_copies = 0
            for i_data in db_info['data']:
                if i_data['host'] == "rucio-catalogue":
                    continue
                if i_data['host'] == hc_host:
                    continue
                if i_data['type'] != d_type:
                    continue
                if d_hash not in i_data['location']:
                    continue
                if i_data['status'] != 'transferred':
                    continue
                valid_disk_copies += 1


            #Determine what to purge
            path_to_purge = None
            pHost = None
            pType = None
            pDict = None
            for i_data in db_info['data']:
                if i_data['host'] != hc_host:
                    continue
                if i_data['type'] != d_type:
                    continue
                if d_hash not in i_data['location']:
                    continue
                path_to_purge = i_data['location']
                pHost = i_data['host']
                pType = i_data['type']
                pDict = i_data

            print( helper.global_dictionary.get("force"))
            #purge without request is ok if:
            # - two valid rucio copies ("OK" for transfer)
            # - one valid rucio copy and one or more valid disk copies
            if ((valid_rucio_copies >= 2) or (valid_rucio_copies == 1 and valid_disk_copies >= 1)) and \
                ( helper.global_dictionary.get("force") == False):

                helper.global_dictionary['logger'].Info("Prepare to purge data path:")
                helper.global_dictionary['logger'].Info(f"> {path_to_purge}")

                #purge disk location
                purge_test = False
                try:
                    shutil.rmtree(path_to_purge)
                    helper.global_dictionary['logger'].Info("> Removed from disk [finished]")
                    purge_test = True
                except IOError as e:
                    helper.global_dictionary['logger'].Error("Unable to purge physical file")

                if purge_test == True:
                    #de-register from database
                    self.db.RemoveDatafield(db_info['_id'], pDict)
                    helper.global_dictionary['logger'].Info("> Removed from runDB [finished]")

            elif helper.global_dictionary.get("force") == True:
                helper.global_dictionary['logger'].Info("WARNING! You activated --force")
                helper.global_dictionary['logger'].Info("Confirm purge of datasets manually!")
                helper.global_dictionary['logger'].Info(f"Location: {path_to_purge}")
                helper.global_dictionary['logger'].Info(f"Type: {pType}")
                helper.global_dictionary['logger'].Info(f"Host: {pHost}")
                helper.global_dictionary['logger'].Info(f"Copies on other disks then {hc_host}: {valid_disk_copies}")
                helper.global_dictionary['logger'].Info(f"Copies in Rucio: {valid_rucio_copies}: {valid_rucio_rses}")

                f_purge = input("Do you with to purge that location (type in yes/no)? ")
                if f_purge == 'yes':
                    # purge disk location
                    purge_test = False
                    try:
                        shutil.rmtree(path_to_purge)
                        helper.global_dictionary['logger'].Info("> Removed from disk [finished]")
                        purge_test = True
                    except IOError as e:
                        helper.global_dictionary['logger'].Error("Unable to purge physical file")

                    if purge_test == True:
                        # de-register from database
                        self.db.RemoveDatafield(db_info['_id'], pDict)
                        helper.global_dictionary['logger'].Info("> Removed from runDB [finished]")

                else:
                    helper.global_dictionary['logger'].Info("You have chosen to not purge data")

            else:
                helper.global_dictionary['logger'].Info("Purging data is rejected!")
                helper.global_dictionary['logger'].Info(f"Location: {path_to_purge}")
                helper.global_dictionary['logger'].Info(f"Type: {pType}")
                helper.global_dictionary['logger'].Info(f"Host: {pHost}")
                helper.global_dictionary['logger'].Info(f"Copies on other disks then {hc_host}: {valid_disk_copies}")
                helper.global_dictionary['logger'].Info(f"Copies in Rucio: {valid_rucio_copies}: {valid_rucio_rses}")


        return 0



#Author: Boris Bauermeister
#Descriptions: A class to manage key words in strings

import datetime

class Keyword():

    def __init__(self):
        self._template = []

    def SetTemplate(self, template):
        #A template is defined as a dictionary with keywords such as:
        # template = {'key1': 'keyword_content1', 'key2': 'keyword_content2', 'key3': 'keyword_content3', ... }
        # Based on this template, this class try to fill up the keyword string(s)
        self._template.append(template)

    def ResetTemplate(self):
        #If necessary you can reset all template dictionaries again
        self._template = []

    def SetString(self, string):
        #this is a string with which needs to be filled up with content from keys.
        pass

    def CompleteTemplate(self, template):
        #Fill a template dictionary with the help from CompleteKeywords()
        for key, val, in template.items():
            val = self.CompleteKeywords(val)
        return template

    def CompleteKeywords(self, val):
        if 'tag_words' not in val:
            return val
        if 'did' not in val:
            return val

        did = val['did']
        #determine the tags which need input information
        for i_tag in val['tag_words']:
            #try to find it in the self._template:
            for i_template in self._template:
                if i_tag in i_template:
                    did= did.replace("{"+i_tag+"}", str(i_template[i_tag]))
            #elif self._eval_db_info(i_tag, db_info) != None:
                #did= did.replace("{"+i_tag+"}", self._eval_db_info(i_tag, db_info))

        val['did'] = did
        return val


    def _eval_db_info(self, key, db_info):
        if key=="detector":
            return db_info['detector']
        elif key=="science_run":
            sr = self._get_science_run( db_info['start'] )
            if sr != -1:
                return sr
            else:
                return "{"+science_run+"}"
        else:
            return None

    def _get_science_run(self, timestamp ):
        #Evaluate science run periods:

        #1) Change from sc0 to sc1:
        dt0to1 = datetime.datetime(2017, 2, 2, 17, 40)

        #Evaluate the according science run number:
        if timestamp <= dt0to1:
            science_run = "000"
        elif timestamp >= dt0to1:
            science_run = "001"
        else:
            science_run = -1
        return science_run



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `admix` package."""


import unittest
from click.testing import CliRunner

from admix import admix
from admix import cli


class TestAdmix(unittest.TestCase):
    """Tests for `admix` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_something(self):
        """Test something."""

    def test_command_line_interface(self):
        """Test the CLI."""
        runner = CliRunner()
        result = runner.invoke(cli.main)
        assert result.exit_code == 0
        assert 'admix.cli.main' in result.output
        help_result = runner.invoke(cli.main, ['--help'])
        assert help_result.exit_code == 0
        assert '--help  Show this message and exit.' in help_result.output



#Author: Boris Bauermeister
#Descriptions: Interface class to varous data format formats (locations, rucio, s3,...)

#import sys
#import datetime
#import os
import json

class Templater():

    def __init__(self):
        self._template_configuration = None
        self._config_path = None
        self._eval_list = {}

    def Config(self, config_path):
        #put in your favourite experimental configuration
        #to describe the rucio cataloge description
        self._config_path = config_path
        self.LoadConfig()


    def LoadConfig(self):
        with open(self._config_path) as f:
            self._template_configuration = json.load(f)

    def FindOccurrences(self, test_string, test_char):
        #https://stackoverflow.com/questions/13009675/find-all-the-occurrences-of-a-character-in-a-string
        return [i for i, letter in enumerate(test_string) if letter == test_char]

    def ExtractTagWords(self, test_string, beg_char, end_char):
        word_list = []
        char_beg = self.FindOccurrences(test_string, beg_char)
        char_end = self.FindOccurrences(test_string, end_char)
        for j in range(len(char_beg)):
            j_char_beg = char_beg[j]
            j_char_end = char_end[j]
            word = test_string[j_char_beg+1:j_char_end]
            if word not in word_list:
                word_list.append(word)
        return word_list

    def ExtractSplit(self, test_string, beg_char, end_char):
        split_list = []

        char_beg = self.FindOccurrences(test_string, beg_char)
        char_end = self.FindOccurrences(test_string, end_char)

        for j in range(len(char_beg)-1):
            j_char_beg = char_beg[j]
            j_char_end = char_end[j]
            if j_char_beg > j_char_end:
                j_char_beg = char_beg[j]
                j_char_end = char_end[j+1]

            split_symbol = test_string[j_char_beg+1:j_char_end]
            split_list.extend(split_symbol)
        return split_list

    def Eval(self, plugin=None, host=None, string=None):
        #Get data types (=keys)

        #Prepare the return:
        self._structure = None
        self._eval_list = {}

        self._types = list(self._template_configuration.keys())

        if len(self._types) == 0:
            print("Upload types are not defined")


        #Get the overall file structure from your configuration file (depends on experiment)

        for i_type in self._types:
            if i_type != plugin:
                continue
            i_levels  = self._template_configuration[i_type]
            if host in list(i_levels.keys()):
                self._structure = i_levels[host]

        if string==None:
            return 0

        if self._structure == None:
            #print("check your config file")
            return 0

        #Evaluate the string according the template:
        tag_words = self.ExtractTagWords(self._structure, "{", "}")
        split_list = self.ExtractSplit(self._structure, "}", "{")


        tag_words = list(reversed(tag_words))
        split_list= reversed(split_list)
        for i, i_split in enumerate(split_list):
            take = string.split(i_split)[-1]
            string = string.replace(i_split+take, "")
            self._eval_list[tag_words[i]]=take
            #print(i, take, string)
        self._eval_list[tag_words[-1]]=string


    def GetTypes(self):
        return self._types

    def GetStructure(self, plugin=None, host=None):
        self.Eval(plugin=plugin, host=host, string=None)
        return self._structure

    def GetTemplateEval(self, plugin=None, host=None, string=None):
        self.Eval(plugin=plugin, host=host, string=string)
        return self._eval_list



# -*- coding: utf-8 -*-
from admix.helper import helper

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater

@Collector
class ClearTransfersMongoDB():

    def __init__(self):
        pass
    def __del__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')

        self.db = ConnectMongoDB()
        self.db.Connect()

        # We want the first and the last run:
        self.gboundary = self.db.GetBoundary()
        self.run_nb_min = self.gboundary['min_number']
        self.run_nb_max = self.gboundary['max_number']
        self.run_ts_min = self.gboundary['min_start_time']
        self.run_ts_max = self.gboundary['max_start_time']

        # Init the Rucio data format evaluator in three steps:
        self.rc_reader = ConfigRucioDataFormat()
        self.rc_reader.Config(helper.get_hostconfig('rucio_template'))

        # This class will evaluate your destinations:
        self.destination = Destination()

        # Since we deal with an experiment, everything is predefine:
        self.exp_temp = Templater()
        self.exp_temp.Config(helper.get_hostconfig()['template'])

        # Init a class to handle keyword strings:
        self.keyw = Keyword()

        # Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()


    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')

        ts_beg = None
        ts_end = None
        if helper.global_dictionary.get('run_numbers') != None:
            # Evaluate terminal input for run number assumption (terminal input == true)
            true_nb_beg, true_nb_end = helper.eval_run_numbers(helper.global_dictionary.get('run_numbers'),
                                                               self.run_nb_min,
                                                               self.run_nb_max)
            # Get the timestamps from the run numbers:
            ts_beg = self.db.FindTimeStamp('number', int(true_nb_beg))
            ts_end = self.db.FindTimeStamp('number', int(true_nb_end))

        elif helper.global_dictionary.get('run_timestamps') != None:
            # Evaluate terminal input for run name assumption
            true_ts_beg, true_ts_end = helper.eval_run_timestamps(helper.global_dictionary.get('run_timestamps'),
                                                                  self.run_ts_min,
                                                                  self.run_ts_max)
            ts_beg = true_ts_beg
            ts_end = true_ts_end

        elif helper.global_dictionary.get('run_timestamps') == None and \
                helper.global_dictionary.get('run_numbers') == None:
            ts_beg = self.run_ts_min
            ts_end = self.run_ts_max
        else:
            helper.global_dictionary['logger'].Error(
                "Check for your input arguments (--select-run-number or --select-run-time")
            exit(1)
            # exection

        # After we know the times:
        helper.global_dictionary['logger'].Info(f"Run between {ts_beg} and {ts_end}")

        #Get your collection of run numbers and run names
        collection = self.db.GetClearance(ts_beg, ts_end)

        #Run through the overview collection:
        for i_run in collection:
            #Extract run number and name from overview collection
            r_name = i_run['name']

            dict_name = {}
            dict_name['date'] = r_name.split("_")[0]
            dict_name['time'] = r_name.split("_")[1]

            r_number = i_run['number']
            #Pull the full run information (according to projection which is pre-defined) by the run name
            db_info = self.db.GetRunByName(r_name)[0]

            #REMOVE AND CHECK FOR UPLOAD DETECTORS (-> config files)
            #ToDo: Make it work with XENONnT later
            if db_info['detector'] != 'tpc':
                continue

            for i_data in db_info['data']:

                #Skip everything what is not host rucio-catalogue
                if i_data.get('host') != "rucio-catalogue":
                    continue
                #Skip everything what has not the requested clearance status
                if i_data.get('status') != "RucioClearance":
                    continue

                #remove the field from the database:
                self.db.RemoveDatafield(db_info['_id'], i_data)

                helper.global_dictionary['logger'].Info("Remove:")
                helper.global_dictionary['logger'].Info(f"Status: {i_data.get('status')}")
                helper.global_dictionary['logger'].Info(f"Location: {i_data.get('location')}")
                helper.global_dictionary['logger'].Info(f"Type: {i_data.get('type')}")
                helper.global_dictionary['logger'].Info(f"for run {r_number}/{db_info['name']} from runDB")




# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper
import time
import shutil

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.utils import make_did

@Collector
class CheckTransfers():
    """
    Using the runDB, it searches for all runs and all data types for which a Rucio rule is ongoing
    (identified by both status specific data.status equal to "transferring".
    For each of them, it checks, by using Rucio API commands, if those rules have been
    succesfully completed. If so, the corresponding data.status is updated as
    "transferred". If all data types are flagged as transferred, then the run itself
    is flagged as "transferred".
    """

    def __init__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')

        #Take all data types categories
        self.NORECORDS_DTYPES = helper.get_hostconfig()['norecords_types']
        self.RAW_RECORDS_DTYPES = helper.get_hostconfig()['raw_records_types']
        self.RECORDS_DTYPES = helper.get_hostconfig()['records_types']

        # Choose which RSE you want upload to
        self.UPLOAD_TO = helper.get_hostconfig()['upload_to']

        #Choose which data type you want to treat
        self.DTYPES = self.NORECORDS_DTYPES + self.RECORDS_DTYPES + self.RAW_RECORDS_DTYPES


        #Define the waiting time (seconds)
        self.waitfor = 60*5

        #Init the runDB
        self.db = ConnectMongoDB()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()
        self.rc.SetProxyTicket("rucio_x509")


    def check_transfers(self):
        cursor = self.db.db.find(
            {'status': 'transferring'},
#            {'number':10161},
            {'number': 1, 'data': 1, 'bootstrax': 1})

        cursor = list(cursor)

        helper.global_dictionary['logger'].Info('Check transfers : checking status of {0} runs'.format(len(cursor)))

        for run in list(cursor):

            # Extracts the correct Event Builder machine who processed this run
            bootstrax = run['bootstrax']
            eb = bootstrax['host'].split('.')[0]

            # for each run, check the status of all REPLICATING rules
            rucio_stati = []
            eb_still_to_be_uploaded = []

            for d in run['data']:
                if d['host'] == 'rucio-catalogue':
#                    if run['number']==7695 and d['status'] == 'stuck':
#                        self.db.db.find_one_and_update({'_id': run['_id'],'data': {'$elemMatch': d}},
#                                                       {'$set': {'data.$.status': 'transferring'}}
#                                                   )                        
                    if d['status'] in ['transferring','error','stuck']:
                        did = d['did']
                        status = self.rc.CheckRule(did, d['location'])
                        if status == 'REPLICATING':
                            rucio_stati.append('transferring')
                        elif status == 'OK':
                            # update database
                            helper.global_dictionary['logger'].Info('Check transfers : Run {0}, data type {1}, location {2}: transferred'.format(run['number'], d['type'],d['location']))
                            self.db.db.find_one_and_update({'_id': run['_id'],'data': {'$elemMatch': d}},
                                                      {'$set': {'data.$.status': 'transferred'}}
                            )
                            rucio_stati.append('transferred')

                        elif status == 'STUCK':
                            self.db.db.find_one_and_update({'_id': run['_id'], 'data': {'$elemMatch': d}},
                                                      {'$set': {'data.$.status': 'stuck'}}
                            )
                            rucio_stati.append('stuck')
                    else:
                        rucio_stati.append(d['status'])


                # search if dtype still has to be uploaded
                if eb in d['host'] and d['type'] in self.DTYPES:
                    if 'status' not in d:
                        eb_still_to_be_uploaded.append(d['type'])
                    else:
                        if d['status'] != "transferred":
                            eb_still_to_be_uploaded.append(d['type'])

            # are there any other rucio rules transferring?
#            print(run['number'],eb_still_to_be_uploaded,rucio_stati)
            if len(rucio_stati) > 0 and all([s == 'transferred' for s in rucio_stati]) and len(eb_still_to_be_uploaded)==0:
                self.db.SetStatus(run['number'], 'transferred')
                helper.global_dictionary['logger'].Info('Check transfers : Run {0} fully transferred'.format(run['number']))





    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')


        # Check transfers
        self.check_transfers()
        

        return 0


    def __del__(self):
        pass



# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper
import time
import shutil

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater
from admix.utils import make_did

@Collector
class UploadFromLNGSSingleThread():

    def __init__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')


        open("/tmp/admix-upload_from_lngs", 'a').close()

        #Take all data types categories
        self.NORECORDS_DTYPES = helper.get_hostconfig()['norecords_types']
        self.RAW_RECORDS_DTYPES = helper.get_hostconfig()['raw_records_types']
        self.RECORDS_DTYPES = helper.get_hostconfig()['records_types']

        # Choose which RSE you want upload to
        self.UPLOAD_TO = helper.get_hostconfig()['upload_to']

        #Choose which data type you want to treat
#        self.DTYPES = self.RAW_RECORDS_DTYPES + self.RECORDS_DTYPES + self.NORECORDS_DTYPES
        self.DTYPES = self.NORECORDS_DTYPES + self.RECORDS_DTYPES + self.RAW_RECORDS_DTYPES


        self.DATADIR = helper.get_hostconfig()['path_data_to_upload']
        self.periodic_check = helper.get_hostconfig()['upload_periodic_check']

        #Init the runDB
        self.db = ConnectMongoDB()

        #We want the first and the last run:
        self.gboundary = self.db.GetBoundary()
        self.run_nb_min = self.gboundary['min_number']
        self.run_nb_max = self.gboundary['max_number']
        self.run_ts_min = self.gboundary['min_start_time']
        self.run_ts_max = self.gboundary['max_start_time']

        #Init the Rucio data format evaluator in three steps:
        self.rc_reader = ConfigRucioDataFormat()
        self.rc_reader.Config(helper.get_hostconfig('rucio_template'))

        #This class will evaluate your destinations:
        self.destination = Destination()

        #Since we deal with an experiment, everything is predefine:
        self.exp_temp = Templater()
        self.exp_temp.Config(helper.get_hostconfig()['template'])

        #Init a class to handle keyword strings:
        self.keyw = Keyword()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()
        self.rc.SetProxyTicket("rucio_x509")
#        print(self.rc.Whoami())



    def find_next_run_to_upload(self):
        cursor = self.db.db.find({'status': 'eb_ready_to_upload', 'bootstrax.state': 'done' }, {'number': 1, 'data': 1})
#        cursor = self.db.db.find({'status': 'uploading'}, {'number': 1, 'data': 1})
        id_run = 0
        min_run = float('inf')

        for run in cursor:
#            if run['number']<8500:
#                continue
#            if run['number'] not in [7155,7156,7218]:
#                 continue
            if run['number'] < min_run:
                min_run = run['number']
                id_run = run['_id']
        return id_run





    def add_rule(self,run_number, dtype, hash, rse, datum=None, lifetime=None, update_db=True):
        did = make_did(run_number, dtype, hash)
        if dtype in self.NORECORDS_DTYPES:
            priority = 1
        else:
            priority = 3
        result = self.rc.AddRule(did, rse, lifetime=lifetime, priority=priority)
        #if result == 1:
        #   return
        helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}: rule added: {2} ---> {3}'.format(run_number,dtype,did,rse))

        if update_db:
            self.db.db.find_one_and_update({'number': run_number},
                                      {'$set': {'status': 'transferring'}}
                                  )

            rucio_rule = self.rc.GetRule(did, rse=rse)
            updated_fields = {'host': "rucio-catalogue",
                         'type': dtype,
                         'location': rse,
                         'lifetime': rucio_rule['expires'],
                         'status': 'transferring',
                         'did': did,
                         'protocol': 'rucio'
            }

            if datum == None:
                data_dict = updated_fields
            else:
                data_dict = datum.copy()
                data_dict.update(updated_fields)

            docid = self.db.db.find_one({'number': run_number}, {'_id': 1})['_id']
            self.db.AddDatafield(docid, data_dict)


    def add_conditional_rule(self,run_number, dtype, hash, from_rse, to_rse, datum=None, lifetime=None, update_db=True):
        did = make_did(run_number, dtype, hash)
        if dtype in self.NORECORDS_DTYPES:
            priority = 1
        else:
            priority = 3
        result = self.rc.AddConditionalRule(did, from_rse, to_rse, lifetime=lifetime, priority=priority)
        #if result == 1:
        #   return
        helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}: conditional rule added: {2} ---> {3}'.format(run_number,dtype,did,to_rse))

        if update_db:
            self.db.db.find_one_and_update({'number': run_number},
                                      {'$set': {'status': 'transferring'}}
                                  )

            rucio_rule = self.rc.GetRule(did, rse=to_rse)
            updated_fields = {'host': "rucio-catalogue",
                         'type': dtype,
                         'location': to_rse,
                         'lifetime': rucio_rule['expires'],
                         'status': 'transferring',
                         'did': did,
                         'protocol': 'rucio'
            }

            if datum == None:
                data_dict = updated_fields
            else:
                data_dict = datum.copy()
                data_dict.update(updated_fields)

            docid = self.db.db.find_one({'number': run_number}, {'_id': 1})['_id']
            self.db.AddDatafield(docid, data_dict)





    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')

#        number = 8113
#        dtype = "raw_records"
#        hash = "rfzvpzj4mf"
#        self.add_conditional_rule(number, dtype, hash, 'UC_OSG_USERDISK', 'CCIN2P3_USERDISK')
#        return 0

        # Get a new run to upload
        id_to_upload = self.find_next_run_to_upload()
        if id_to_upload == 0:
             helper.global_dictionary['logger'].Info('No run available to upload')
             return 0
        run = self.db.db.find_one({'_id': id_to_upload}, {'number': 1, 'data': 1, 'bootstrax': 1})

        # Gets run number
        number = run['number']

        # Books the run by setting its status to the PID of the process
        PID = str(os.getpid())
        self.db.SetStatus(number, PID)
#        self.db.SetStatus(number, 'uploading')

        # Wait for 10 seconds
        time.sleep(10)

        # Then check if status is still equal to the same PID
        run = self.db.db.find_one({'_id': id_to_upload}, {'status': 1, 'number': 1, 'data': 1, 'bootstrax': 1})
        if run['status'] != PID:
            helper.global_dictionary['logger'].Info('Run {0} already taken by another process'.format(number))
            return 0

        # If yes, then book it by setting its status to 'uploading'
        self.db.SetStatus(number, 'uploading')


        # Extracts the correct Event Builder machine who processed this run
        bootstrax = run['bootstrax']
        eb = bootstrax['host'].split('.')[0]

        # Performs upload on selected run
        helper.global_dictionary['logger'].Info('Uploading run {0} from {1}'.format(number,eb))

        # loop on all data types we want to upload
        for dtype in self.DTYPES:
            helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}: uploading'.format(number,dtype))

            # get the datum for this datatype
            datum = None
            in_rucio_upload_rse = False
            in_rucio_somewhere_else = False
            for d in run['data']:
                if d['type'] == dtype and eb in d['host']:
                    datum = d

                if d['type'] == dtype and d['host'] == 'rucio-catalogue' and d['location'] == self.UPLOAD_TO:
                    in_rucio_upload_rse = True

                if d['type'] == dtype and d['host'] == 'rucio-catalogue' and d['location'] != self.UPLOAD_TO:
                    in_rucio_somewhere_else = True

            if datum is None:
                helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}: not found'.format(number,dtype))
                continue

            file = datum['location'].split('/')[-1]

            hash = file.split('-')[-1]

            upload_path = os.path.join(self.DATADIR, eb, file)

            # create a DID to upload
            did = make_did(number, dtype, hash)

            # check if a rule already exists for this DID on LNGS
            rucio_rule = self.rc.GetRule(upload_structure=did, rse=self.UPLOAD_TO)
#            helper.global_dictionary['logger'].Info('It was already in Rucio : {0}'.format(in_rucio_upload_rse))
#            helper.global_dictionary['logger'].Info('Rucio rule : {0}'.format(rucio_rule))
#            print("Did: ",did)
#            print("Upload path: ",upload_path)

            if in_rucio_somewhere_else:
                helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}: cannot upload it because it is already on Rucio somewhere else'.format(number,dtype))
                continue

            # if not in rucio already (no matter where) and no rule exists, upload into rucio
            if not in_rucio_upload_rse and not rucio_rule['exists']:
                result = self.rc.Upload(did,
                                        upload_path,
                                        self.UPLOAD_TO,
                                        lifetime=None)
                helper.global_dictionary['logger'].Info('\t==> Run {0}, data type {1}: uploaded ({2})'.format(number,dtype,did))
                
                # get the status of this new upload rule
                rucio_rule = self.rc.GetRule(upload_structure=did, rse=self.UPLOAD_TO)

                # if the rule status is OK, then update the DB
                if rucio_rule['state'] == 'OK':
                    data_dict = datum.copy()
                    data_dict.update({'host': "rucio-catalogue",
                                 'type': dtype,
                                 'location': self.UPLOAD_TO,
                                 'lifetime': rucio_rule['expires'],
                                 'status': 'transferred',
                                 'did': did,
                                 'protocol': 'rucio'
                             })
                    self.db.AddDatafield(run['_id'], data_dict)

            # set a rule to ship data on GRID
            if rucio_rule['state'] == 'OK':
                if dtype in self.NORECORDS_DTYPES:
                    self.add_rule(number, dtype, hash, 'UC_DALI_USERDISK',datum=datum)
                    self.add_conditional_rule(number, dtype, hash, 'UC_DALI_USERDISK', 'UC_OSG_USERDISK',datum=datum)
                else:
                    self.add_rule(number, dtype, hash, 'UC_OSG_USERDISK',datum=datum)
                    if dtype in self.RECORDS_DTYPES:
                        self.add_conditional_rule(number, dtype, hash, 'UC_OSG_USERDISK', 'UC_DALI_USERDISK',datum=datum)
                    self.add_conditional_rule(number, dtype, hash, 'UC_OSG_USERDISK', 'CCIN2P3_USERDISK',datum=datum)


        return 0


    def __del__(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

# Get requirements from requirements.txt, stripping the version tags
with open('requirements.txt') as f:
    requirements = [r.split('/')[-1] if r.startswith('git+') else r for r in f.read().splitlines()]

setup_requirements = [
    # TODO(XeBoris): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='xe-admix',
    version='0.2.0',
    description="advanced Data Management In Xenon (aDMIX)",
    long_description=readme + '\n\n' + history,
    url='https://github.com/XENON1T/admix',
    packages=find_packages(include=['admix',
                                    'admix.interfaces',
                                    'admix.tasks',
                                    'admix.helper',
                                    'admix.utils'
                                    ]),
    package_data={'admix.helper': ['defunc_'],
                  'admix': ['config/*.*', 'config/rucio_cli/*.*']},
    include_package_data=True,
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'admix-version=admix.admix:version',
            'admix=admix.admix:your_admix',
            'admix-download=admix.download:main',
            'admix-showrun=admix.showrun:main',
            'admix-showcontexts=admix.showcontexts:main'
        ]
    },
    license="BSD license",
    zip_safe=False,
    keywords='admix',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6'
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



import argparse
import logging
import numpy as np
import os
import time

from admix.helper.logger import Logger
import admix.helper.helper as helper
from admix import DEFAULT_CONFIG, __version__

from admix.helper.decorator import NameCollector, ClassCollector
from admix.tasks.example_task import RunExampleTask
from admix.tasks.upload_with_mongodb import UploadMongoDB
from admix.tasks.update_runDB import UpdateRunDBMongoDB
from admix.tasks.init_transfers_with_mongodb import InitTransfersMongoDB
from admix.tasks.download_with_mongodb import DownloadMongoDB
from admix.tasks.clear_transfers_with_mongdb import ClearTransfersMongoDB
from admix.tasks.purge_with_mongodb import PurgeMongoDB
from admix.tasks.upload_from_lngs import UploadFromLNGS
from admix.tasks.fix_upload import FixUpload
from admix.tasks.clean_eb import CleanEB
from admix.tasks.upload_from_lngs_single_thread import UploadFromLNGSSingleThread
from admix.tasks.upload import Upload
from admix.tasks.check_transfers import CheckTransfers
from admix.tasks.move_data_to_rse import MoveDataToRSE
from admix.tasks.monitor_run import MonitorRun
from utilix.config import Config

def version():
    print(__version__)

def your_admix():
    print("advanced Data Management in XENON")

    parser = argparse.ArgumentParser(description="Run your favourite aDMIX")

    config = Config()


    # From here the input depends on the usage of the command
    # Add modules here:
    parser.add_argument('task', nargs="?", default="default",
                        help="Select an aDMIX task")
    # Add arguments for the process manager:
    parser.add_argument('--admix-config', dest="admix_config", type=str, default=config.get('Admix','config_file'),
                        help="Load your host configuration")
    parser.add_argument('--no-update', dest='no_update', action='store_false',
                        help="Add this option to prevent aDMIX updating the Xenon database")
    parser.add_argument('--once', dest='once', action='store_true',
                        help="Run aDMIX only once")
    parser.add_argument('--high', dest='high', action='store_true',
                        help="Treat only high level data types")
    parser.add_argument('--low', dest='low', action='store_true',
                        help="Treat only low level data types")
    # Add arguments for the individual tasks:
    parser.add_argument('--select-run-numbers', dest='select_run_numbers', type=str,
                        help="Select a range of runs (xxxx1 or xxxx1-xxxx2 or xxxx1-xxxx2,xxxx4-xxxx6)")
    parser.add_argument('--select-run-times', dest='select_run_times', type=str,
                        help="Select a range of runs by timestamps <Date>_<Time>-<Date>_<Time>")
    parser.add_argument('--source', nargs='*', dest='source', type=str,
                        help="Select data according to a certain source(s)")
    parser.add_argument('--type', nargs='*', dest='type', type=str,
                        help="Select data according to a certain type")
    parser.add_argument('--hash', nargs='*', dest='hash', type=str,
                        help="Select data according to a certain hash")
    parser.add_argument('--tag', nargs='*', dest='tag', type=str,
                        help="Select data according to a certain tag(s)")
    parser.add_argument('--destination', dest='destination', type=str,
                        help="Add a destination from ")
    parser.add_argument('--rse', dest='rse', type=str,
                        help="Select your RSE from where to download data")
    parser.add_argument('--lifetime', dest='lifetime', type=str,
                        help="Select your RSE from where to download data")
    parser.add_argument('--force', default=False, action="store_true",
                        help="Enforce your action. Be aware of the application!")
    parser.add_argument('--sleep-time', dest='sleep_time', type=int,
                        help="Time to wait before running again the task")
    args = parser.parse_args()

    #We make the individual arguments global available right after aDMIX starts:
    if args.select_run_numbers != None and args.select_run_times == None:
        helper.make_global("run_numbers", args.select_run_numbers)
    if args.select_run_times != None and args.select_run_numbers == None:
        helper.make_global("run_timestamps", args.select_run_times)

    helper.make_global("admix_config", os.path.abspath(args.admix_config))
    helper.make_global("no_db_update", args.no_update)
    helper.make_global("source", args.source)
    helper.make_global("tag", args.tag)
    helper.make_global("type", args.type)
    helper.make_global("hash", args.hash)
    helper.make_global("force", args.force)
    helper.make_global("high", args.high)
    helper.make_global("low", args.low)

    if args.sleep_time != None:
        helper.make_global("sleep_time", args.sleep_time)
    else:
        helper.make_global("sleep_time",helper.get_hostconfig()['sleep_time'])

    #Pre tests:
    # admix host configuration must match the hostname:
    if helper.get_hostconfig()['hostname'] != helper.get_hostname():
        print("admix configuration file for %s" % helper.get_hostconfig()['hostname'])
        print(helper.get_hostname())
        print("You are at {0}".format( helper.get_hostname()))
        exit()

#    helper.functdef()

    #Setup the logger in a very basic modi
    lg = Logger(logpath=helper.get_hostconfig()['log_path'],
                loglevel=logging.DEBUG)
    lg.Info("-----------------------------------------")
    lg.Info("aDMIX - advanced Data Management in XENON")
    helper.make_global("logger", lg)

    #Determine which tasks are addressed:
    # - if it comes from args.task use it, nevertheless what is defined in hostconfig("task")
    # - if args.task == default use hostconfig("task") information
    task_list = []
    if args.task == "default":
        task_list.extend(helper.get_hostconfig("task"))
    else:
        task_list = [args.task]

    #test if the list of tasks is available from the decorator
    task_test = [True if i_task in NameCollector else False for i_task in task_list]
    task_list = np.array(task_list)[task_test]

    if len(task_list) == 0:
        print("Select a task from this list:")
        for i_task in NameCollector:
            print("  <> {0}".format(i_task))
        print("or adjust the 'task' field in your configuration")
        print("file: {0}".format(helper.global_dictionary["admix_config"]))
        exit()

    #Loop over the inizialization of all classes
    for i_task in task_list:
        ClassCollector[i_task].init()

    #Go for the loop
    while True:

        for i_task in task_list:
            ClassCollector[i_task].run()


        if args.once == True:
            break

        if os.path.exists("/tmp/admix-stop"):
            print("Exiting because of the presence of /tmp/admix-stop file")
            break

        wait_time = helper.global_dictionary['sleep_time']
        if "CheckTransfers" in task_list or "CleanEB" in task_list:
            wait_time = 120

        print('Waiting for {0} seconds'.format(wait_time))
        print("You can safely CTRL-C now if you need to stop me")
        try:
            time.sleep(wait_time)

        except KeyboardInterrupt:
            break


